#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype Inferential Type-hint Engine (BITE) callable type hint inferrers**
(i.e., lower-level functions dynamically subscripted type hints describing
callable objects).
'''

# ....................{ TODO                               }....................
#FIXME: We can and should do better. Currently, infer_hint_callable() reduces
#all unannotated variadic positional arguments (e.g., "*args") and annotated
#variadic positional arguments (e.g., "*args: int") to simply
#"Concatenate[...]", basically. Previously, we thought that was the best we
#could do. However, it turns out that PEP 646 under Python >= 3.11 provides an
#obscure means of representing variadic positional arguments with *UNPACKED
#TUPLE TYPE HINTS*. Notably:
#    The behavior of a Callable containing an unpacked item, whether the item is
#    a TypeVarTuple or a tuple type, is to treat the elements as if they were
#    the type for *args. So, Callable[[*Ts], None] is treated as the type of the
#    function:
#        def foo(*args: *Ts) -> None: ...
#
#    Callable[[int, *Ts, T], Tuple[T, *Ts]] is treated as the type of the
#    function:
#        def foo(*args: *Tuple[int, *Ts, T]) -> Tuple[T, *Ts]: ...
#
#Given the above, it *SHOULD* theoretically follow that the type hint inferred
#from a callable with signature:
#    def muh_func(*args: int) -> None:
#
#...is this "Callable[...]" type hint:
#    Callable[[*tuple[int]], None]
#
#We know. Looks weird, but that's what a casual reading of the above section
#explicitly suggests. Let's test that theory against mypy, please.
#FIXME: *HMM.* Looks like mypy currently fails to type-check this properly. mypy
#assorts that the following code snippet passes. Surely that isn't right,
#though?
#    from collections.abc import Callable
#
#    def muh_func(*args: int) -> None: pass
#    def moo_func(*args: float) -> None: pass
#    def ugh(glugh: Callable[[*tuple[int]], None]) -> None: pass
#
#    ugh(muh_func)  # <-- this passes. GOOD!
#    ugh(moo_func)  # <-- this passes, too. BAD!
#
#Why does mypy pass both? Is mypy bugged? Is our reading of PEP 646 bugged? Does
#mypy simply fail to fully support PEP 646? That would kinda explain why nobody
#has bothered @beartype about supporting PEP 646, frankly. We're currently
#inclined to believe that mypy has yet to implement this. Surely mypy should be
#saying *SOMETHING* about that case, even if that something is to just notify us
#of how "creative" we are. Let's check back sometime in 2025, please.

#    >>> from beartype.bite import infer_hint
#
#    # Ideally, this...
#    >>> infer_hint(lambda: None, True)
#    collections.abc.Callable
#
#    # ...should instead resemble this.
#    >>> infer_hint(lambda: None, True)
#    collections.abc.Callable[[], object]

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Callable,
    Optional,
)
from beartype._cave._cavefast import (
    HintPep612ParamSpecArgType,
    HintPep612ParamSpecType,
)
from beartype._data.func.datafuncarg import ARG_NAME_RETURN
from beartype._data.typing.datatypingport import (
    Hint,
    ListHints,
)
from beartype._data.hint.sign.datahintsigns import (
    HintSignParamSpecArgs,
    HintSignParamSpecKwargs,
)
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.func.arg.utilfuncargiter import (
    ArgKind,
    ArgMandatory,
    iter_func_args,
)
from beartype._util.func.utilfunctest import is_func_codeobjable
from beartype._util.func.utilfuncwrap import unwrap_func_all_isomorphic
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none
from beartype._util.hint.pep.proposal.pep612 import (
    get_hint_pep612_paramspec,
    make_hint_pep612_concatenate_list_or_none,
)
from beartype._util.hint.pep.proposal.pep649 import (
    get_pep649_hintable_annotations)
from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11
from collections.abc import (
    Callable as CallableABC,
)

# ....................{ INFERERS                           }....................
#FIXME: Unit test us up, please.
#FIXME: Revise docstring up, please.
@callable_cached
def infer_hint_callable(func: CallableABC) -> Hint:
    '''
    **Callable type hint** (i.e., :class:`collections.abc.Callable` protocol
    possibly subscripted by two or more child type hints describing the
    parameters and return) validating the passed callable.

    Specifically, this function (in order):

    #. If the passed callable is non-pure-Python *and*:

       * If the passed callable wraps a lower-level pure-Python callable (e.g.,
         via the standard :func:`functools.wraps` decorator), then the passed
         callable is iteratively unwrapped until obtaining the lowest-level
         pure-Python callable that is *not* such a wrapper.
       * Else, the unsubscripted :class:`collections.abc.Callable` protocol is
         returned as is.

    #. If that callable is unannotated, the unsubscripted
       :class:`collections.abc.Callable` protocol is returned as is.
    #. Else, that callable is annotated by one or more parameter and/or return
       child type hints. If that callable accepts *no* parameters, that callable
       *must* have been annotated by only a return child type hint. In this
       case, the :pep:`585`-compliant type hint ``collections.abc.Callable[...,
       {hint_return}]`` is returned.
    #. Else, that callable accepts one or more parameters. If those parameters
       are all unannotated, the :pep:`585`-compliant type hint
       ``collections.abc.Callable[..., {hint_return}]`` is again returned.
    #. Else, one or more of those parameters are annotated. If that callable
       accepts only mandatory positional-only and/or flexible parameters, the
       :pep:`585`-compliant type hint
       ``collections.abc.Callable[[{hint_arg_1}, ???, {hint_arg_N}],
       {hint_return}]`` is returned.
    #. Else, that callable accepts one or more of the following:

       * An optional positional-only parameter.
       * An optional flexible parameter.
       * A variadic positional parameter.
       * A keyword-only parameter.
       * A variadic keyword parameter.

       Several distinct cases now arise:

       * If that callable accepts a variadic positional parameter annotated by
         the :pep:`612`-compliant ``P.args`` parameter child type hint *and* a
         corresponding variadic keyword parameter annotated by the
         :pep:`612`-compliant ``P.kwargs`` parameter child type hint for some
         :pep:`612`-compliant **parameter specification** (i.e.,
         :obj:`typing.ParamSpec` object) ``P`` *and*:

         * No other parameters, the :pep:`612`-compliant type hint
           ``collections.abc.Callable[P, {hint_return}]`` is returned.
         * Only one or more mandatory positional-only and/or flexible
           parameters, the :pep:`612`-compliant type hint
           ``collections.abc.Callable[typing.Concatenate[{hint_arg_1}, ???,
           {hint_arg_N}, P]], {hint_return}]`` is returned.
         * Else, that callable accepts one or more parameters that *cannot* be
           annotated under any existing PEP standard. In this case, these
           unsupported parameters *must* be ignored by replacing the trailing
           ``P`` subscripting the prior ``typing.Concatenate[???]`` child type
           hint with an **ellipsis** (i.e., ``...`` singleton). Although
           undocumented, at least mypy_ and possibly other static type-checkers
           (e.g., pyright_) have extended :pep:`612` for unknown reasons to
           permit this replacement. Interestingly, this undocumented behaviour
           does *not* support both a parameter specification and an ellipsis;
           this behaviour only supports either a parameter specification *or* an
           ellipsis. While the ellipsis is necessary here, the parameter
           specification is not and *must* thus be dropped. Let ``N`` be the
           0-based index of the last mandatory positional-only and/or flexible
           parameter in the signature of that callable. Then, in this case, the
           ludicrous :pep:`612`-compliant type hint
           ``collections.abc.Callable[typing.Concatenate[{hint_arg_1}, ???,
           {hint_arg_N}, ...]], {hint_return}]`` is returned.

       * Else, that callable again accepts one or more parameters that *cannot*
         be annotated under any existing PEP standard. In this case, these
         unsupported parameters *must* be ignored by abusing the
         :pep:`612`-compliant :obj:`typing.Concatenate` type hint factory.
         Although the child parameter list of :pep:`585`-compliant
         ``collections.abc.Callable[[{hint_args}], {hint_return}]`` type hints
         does *not* support an ellipsis, the child parameter list of
         :pep:`612`-compliant
         ``collections.abc.Callable[typing.Concatenate[{hint_args}],
         {hint_return}]`` type hints does. In theory, :pep:`612` should *not*
         apply here, as that callable accepts *no* :pep:`612`-compliant
         parameter specification. In practice, :pep:`612` can be abused to
         support otherwise unsupportable parameters. Again, let ``N`` be the
         0-based index of the last mandatory positional-only and/or flexible
         parameter in the signature of that callable. Then, in this case, the
         ludicrous :pep:`612`-compliant type hint
         ``collections.abc.Callable[typing.Concatenate[{hint_arg_1}, ???,
         {hint_arg_N}, ...]], {hint_return}]`` is returned.

    Although admittedly imperfect, this inference strategy nonetheless supports
    a *much* broader set of callable signatures with *much* narrower type hints
    than existing PEP standards would superficially appear to support.

    This function is memoized for efficiency.

    Parameters
    ----------
    func : CallableABC
        Callable to infer a type hint from.

    Returns
    -------
    Hint
        Callable type hint validating this callable.

    See Also
    --------
    https://stackoverflow.com/a/77467213/2809027
        StackOverflow answer inspiring this implementation.
    '''

    # ....................{ LOCALS                         }....................
    # Preserve the passed callable under a different name for subsequent reuse.
    func_wrapper = func

    # If the passed callable isomorphically wraps another lower-level callable
    # (i.e., if the former is a @functools.wraps()-decorated callable accepting
    # *ONLY* "*args, **kwargs" parameters), reduce the former to the latter.
    func = unwrap_func_all_isomorphic(func_wrapper)

    #FIXME: *HANDLE THIS.* Do so in a similar manner to that of the
    #beartype._decor.decornontype.beartype_pseudofunc() function, please.
    #Namely, attempt to see whether the "func.__call__" instance variable exists
    #and is a pure-Python callable. If so, replace "func" with that: e.g.,
    #    func = getattr(func, '__call__', None)
    #    if func is None:
    #        return obj.__class__  # <-- don't return "Callable" here, because.
    #
    #Even then, we should embed this "Callable[...]" type hint inside an
    #"Annotated[...]" parent type hint ensuring that only objects of the passed
    #object's type are matched. See similar logic in the "infercollectionsabc"
    #submodule: e.g.,
    #    return Annotated[hint, IsInstance[obj.__class__]]

    # If this unwrapped callable is *NOT* pure-Python, this is a pseudo-callable
    # (i.e., arbitrary pure-Python *OR* C-based object whose class defines the
    # __call__() dunder method enabling this object to be called like a standard
    # callable). In this case, trivially return the unsubscripted
    # "collections.abc.Callable" protocol.
    if not is_func_codeobjable(func):
        return Callable
    # Else, this callable is pure-Python.

    # Hint to be returned, defaulting to the unsubscripted
    # "collections.abc.Callable" protocol.
    hint: Hint = Callable

    # Dictionary mapping from the name of each annotated parameter accepted by
    # that unwrapped callable to the type hint annotating that parameter.
    #
    # Note that:
    # * The functools.update_wrapper() function underlying the
    #   @functools.wrap decorator underlying all sane decorators propagates this
    #   dictionary from lower-level wrappees to higher-level wrappers by
    #   default. We intentionally access the annotations dictionary of this
    #   higher-level wrapper, which *SHOULD* be the superset of that of this
    #   lower-level wrappee (and thus more reflective of reality).
    # * Even if that callable is unannotated (i.e., this dictionary is empty),
    #   we explicitly do *NOT* silently reduce to a noop by returning the
    #   unsubscripted "collections.abc.Callable" protocol. Why not? Because this
    #   function infers more than merely the type hints annotating callable
    #   parameters and returns; this function also infers the *NUMBER* of
    #   parameters accepted by that callable: e.g.,
    #       # This function correctly infers that a lambda function accepting
    #       # *NO* parameters should be annotated as such.
    #       >>> infer_hint(lambda: None)
    #       collections.abc.Callable[[], object]
    pith_name_to_hint = get_pep649_hintable_annotations(func_wrapper)

    # dict.get() method bound to this dictionary, localized for efficiency.
    pith_name_to_hint_get = pith_name_to_hint.get

    # Child hint annotating the parameter(s) of that callable.
    #
    # This hint is initially defined as the list of all child hints annotating
    # *ONLY* the mandatory positional-only and mandatory flexible parameters of
    # that callable if that callable accepts such parameters *OR* the empty list
    # otherwise. This list is subsequently used to synthesize the lead child
    # hints subscripting the parent PEP 612-compliant "typing.Concatenate[...]"
    # hint, which then becomes the first child hint subscripting the returned
    # root PEP 585-compliant "typing.Callable[...]" hint.
    #
    # This initial list may be subsequently replaced with a more appropriate
    # hint suitable for use as the first child hint subscripting the returned
    # PEP 585-compliant "typing.Callable[...]" hint. In short: non-trivial af.
    hint_args: object = []

    # Child hint annotating the return of that callable, defined as either:
    # * If this return is annotated, the hint annotating this return.
    # * Else, by process of elimination, one or more parameters of this callable
    #   are annotated instead. Logic below synthesizes the parent hint to be
    #   returned by subscripting the "collections.abc.Callable" hint factory
    #   with the list of these parameter child hints. PEP 484 makes *NO*
    #   provision for subscripting this factory with only
    #   parameter child hints (and no return child hint); instead, PEP 484
    #   requires this factory *ALWAYS* be subscripted by both parameter and
    #   return child hints. Ergo, a return child hint is *ALWAYS* necessary.
    #   This being the case, we default to the ignorable "object" supertype.
    hint_return = pith_name_to_hint_get(ARG_NAME_RETURN, object)

    # True only if that callable accepts one or more parameters.
    is_args = False

    # True only if one or more parameters of that callable *cannot* be
    # annotated as parameter child type hints of a parent
    # "Callable[[...], Any]" type hint.
    #
    # See the docstring for detailed discussion of which kinds of parameters can
    # and cannot be annotated as parameter child type hints.
    is_arg_unhintable = False

    # PEP 612-compliant parameter specification variadic positional parameter
    # instance variable (e.g., "*args: P.args") if that callable accepts a
    # variadic positional parameter annotated by such a variable *OR* "None".
    pep612_paramspec_args: Optional[HintPep612ParamSpecArgType] = None

    # PEP 612-compliant parameter specification that is the parent of both
    # variadic positional and keyword parameter instance variables annotating
    # the variadic positional and keyword parameters accepted by that callable
    # (e.g., the "P" in "*args: P.args, **kwargs: P.kwargs") if that callable
    # accepts such parameters *OR* "None" otherwise.
    pep612_paramspec: Optional[HintPep612ParamSpecType] = None

    # ....................{ SEARCH                         }....................
    # For the kind, name, and default value of of each parameter accepted by
    # that callable (in declaration order)...
    for arg_kind, arg_name, arg_default in iter_func_args(
        # Possibly lowest-level wrappee underlying the passed possibly
        # higher-level wrapper. The latter typically fails to convey the same
        # callable metadata conveyed by the former -- including the names and
        # kinds of parameters accepted by the possibly unwrapped callable. This
        # renders the latter mostly useless for our purposes.
        func=func,
        # Avoid inefficiently attempting to re-unwrap this wrappee.
        is_unwrap=False,
    ):
        # Child hint annotating this parameter if any *OR* the root "object"
        # superclass otherwise.
        hint_arg: Hint = pith_name_to_hint_get(arg_name, object)

        # Note that that callable accepts one or more parameters.
        is_args = True

        # If this is a keyword-only parameter, this parameter is unsupported by
        # any existing PEP standard. Record this and halt iteration.
        if arg_kind is ArgKind.KEYWORD_ONLY:
            is_arg_unhintable = True
            break
        # Else, this is *NOT* a keyword-only parameter.
        #
        # If this is a variadic positional parameter...
        elif arg_kind is ArgKind.VARIADIC_POSITIONAL:
            # If this parameter is annotated...
            if hint_arg is not object:
                # Sign uniquely identifying this hint if this hint is
                # PEP-compliant *OR* "None" (i.e., if this hint is a
                # PEP-noncompliant type).
                hint_arg_sign = get_hint_pep_sign_or_none(hint_arg)

                # If this variadic positional parameter is annotated by a PEP
                # 612-compliant parameter specification variadic positional
                # parameter instance variable (e.g., resembling "*args:
                # P.args"), record this and continue to the next parameter.
                if hint_arg_sign is HintSignParamSpecArgs:
                    pep612_paramspec_args = hint_arg  # pyright: ignore
                    continue
                # Else, this variadic positional parameter is annotated by a
                # type hint unsupported by any existing PEP standard.
            # Else, this parameter is unannotated.

            # Else, this variadic positional parameter is either unannotated
            # *OR* annotated by a type hint unsupported by any existing PEP
            # standard. In either case, record this and halt iteration.
            is_arg_unhintable = True
            break
        # Else, this is *NOT* a variadic positional parameter.
        #
        # If this is a variadic keyword parameter...
        elif arg_kind is ArgKind.VARIADIC_KEYWORD:
            # If this parameter is annotated...
            if hint_arg is not object:
                # Sign uniquely identifying this hint if this hint is
                # PEP-compliant *OR* "None" (i.e., if this hint is a
                # PEP-noncompliant type).
                hint_arg_sign = get_hint_pep_sign_or_none(hint_arg)

                # If this variadic keyword parameter is annotated by a PEP
                # 612-compliant parameter specification variadic keyword
                # parameter instance variable (e.g., resembling "**kwargs:
                # P.kwargs")..., record this and continue to the next parameter.
                if hint_arg_sign is HintSignParamSpecKwargs:
                    # If that callable also accepts a variadic positional
                    # parameter annotated by a corresponding PEP 612-compliant
                    # parameter specification variadic positional parameter
                    # instance variable...
                    if pep612_paramspec_args:
                        # PEP 612-compliant parent parameter specifications
                        # containing this pair of instance variables.
                        pep612_paramspec_args_paramspec = (
                            get_hint_pep612_paramspec(
                                pep612_paramspec_args))
                        pep612_paramspec_kwargs_paramspec = (
                            get_hint_pep612_paramspec(hint_arg))  # pyright: ignore

                        # If this pair of instance variables derive from the
                        # same PEP 612-compliant parent parameter specification,
                        # record this and halt iteration. By definition, *NO*
                        # parameters follow this parameter; the variadic keyword
                        # parameter is *ALWAYS* the last possible parameter in
                        # any parameter list.
                        if (
                            pep612_paramspec_args_paramspec is
                            pep612_paramspec_kwargs_paramspec
                        ):
                            pep612_paramspec = (
                                pep612_paramspec_args_paramspec)
                            break
                        # Else, this pair of instance variables derive from
                        # different PEP 612-compliant parent parameter
                        # specifications. This is horrible. Who would do
                        # something horrible like this? Somebody horrible.
                    # Else, that callable fails to accept such a variadic
                    # positional parameter.
                # Else, this variadic keyword parameter is annotated by a type
                # hint unsupported by any existing PEP standard.
            # Else, this parameter is unannotated.

            # Else, this variadic keyword parameter is either unannotated *OR*
            # annotated by a type hint unsupported by any existing PEP standard.
            # In either case, record this and halt iteration.
            is_arg_unhintable = True
            break
        # Else, this is *NOT* a variadic keyword parameter. By process of
        # elimination, this *MUST* be either a positional-only or flexible
        # parameter.
        #
        # If this is an optional positional-only or flexible parameter, this
        # parameter is unsupported by any existing PEP standard. Record this and
        # halt iteration.
        elif arg_default is not ArgMandatory:
            is_arg_unhintable = True
            break
        # Else, this is either a mandatory positional-only *OR* mandatory
        # flexible parameter. In either case, append the child hint annotating
        # this mandatory parameter to the list of all child parameter hints.
        else:
            # print(f'arg "{arg_name}": Found Concatenate[...] child hint {hint_arg}...')
            hint_args.append(hint_arg)  # type: ignore[attr-defined]

    # ....................{ SUBSCRIPT                      }....................
    # If...
    if (
        # That callable accepts a variadic positional parameter annotated by a
        # PEP 612-compliant parameter specification variadic positional
        # parameter instance variable (e.g., resembling "*args: P.args")
        # *AND*...
        pep612_paramspec_args is not None and
        # That callable accepts *NO* corresponding variadic keyword parameter...
        pep612_paramspec is None
    ):
        # Record this variadic positional parameter to be unsupported by any
        # existing PEP standard.
        is_arg_unhintable = True

    # If that callable accepts *NO* parameters, preserve the list of all child
    # parameter hints as the empty list.
    if not is_args:
        pass
    # Else, that callable accepts one or more parameters.
    #
    # If that callable accepts one or more unsupported parameters...
    elif is_arg_unhintable:
        # If the active Python interpreter targets Python >= 3.11, then the PEP
        # 612-compliant "typing(|_extensions).Concatenate" hint factory is
        # subscriptable by a trailing ellipsis. In this case...
        if IS_PYTHON_AT_LEAST_3_11:
            # Generalize the list of all child parameter hints to this factory
            # subscripted by all child hints annotating all mandatory
            # positional-only and flexible parameters accepted by that callable
            # suffixed by the ellipsis ignoring all remaining unsupported
            # parameters.
            #
            # See the docstring for detailed discussion of this behaviour.
            hint_args = make_hint_pep612_concatenate_list_or_none(
                hints_child_first=hint_args, hint_child_last=...)  # type: ignore[arg-type]
            # print(f'PEP 612 Concatenate list: {repr(hint_args)}')

            # If this factory is unimportable, we sadly have *NO* recourse but
            # to silently ignore *ALL* parameters.
            if hint_args is None:
                hint_args = ...
            # Else, this factory is importable. In this case, preserve this hint
            # as is.
        # Else, the active Python interpreter targets Python < 3.11. In this
        # case, the PEP 612-compliant "typing(|_extensions).Concatenate" hint
        # factory is *NOT* subscriptable by a trailing ellipsis. Attempting to
        # do so raises:
        #     TypeError: The last parameter to Concatenate should be a ParamSpec
        #     variable.
        #
        # In this case, we sadly have *NO* recourse but to silently ignore *ALL*
        # parameters.
        else:
            hint_args = ...
    # Else, that callable accepts *NO* unsupported parameters.
    #
    # If that callable accepts a PEP 612-compliant parameter specification...
    elif pep612_paramspec is not None:
        # If that callable accepts one or more mandatory positional-only or
        # flexible parameters...
        if hint_args:
            # Generalize the list of all child parameter hints to the PEP
            # 612-compliant "Concatenate[...]"  hint factory subscripted by all
            # child hints annotating all mandatory positional-only and flexible
            # parameters accepted by that callable suffixed by the this
            # parameter specification.
            hint_args = make_hint_pep612_concatenate_list_or_none(
                hints_child_first=hint_args, hint_child_last=pep612_paramspec)  # type: ignore[arg-type]

            # If this factory is unimportable, we sadly have *NO* recourse but
            # to silently ignore *ALL* parameters.
            if hint_args is None:
                hint_args = ...
            # Else, this factory is importable. In this case, preserve this hint
            # as is.
        # Else, that callable accepts *NO* mandatory positional-only or flexible
        # parameters. In this case, set the list of all child parameter hints to
        # this parameter specification as is.
        else:
            hint_args = pep612_paramspec
    # Else, that callable accepts *NO* PEP 612-compliant parameter
    # specification. In this case, that callable accepts *ONLY* one or more
    # mandatory positional-only and/or flexible parameters. In this case,
    # preserve this list of all child parameter hints as is.

    # ....................{ RETURN                         }....................
    # Callable hint to be returned, defined as either...
    hint = (
        # The unsubscripted "collections.abc.Callable" protocol if...
        Callable
        if (
            # The list of all child parameter hints is the ellipsis (signifying
            # all parameters to be ignorable) *AND*...
            hint_args is ... and
            # The child return hint is the ignorable "object" superclass.
            hint_return is object
        ) else
        # Else, the "collections.abc.Callable" type hint factory subscripted by
        # these unignorable child parameter and return type hints.
        Callable[hint_args, hint_return]  # pyright: ignore
    )

    # Return this hint.
    return hint



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **third-party module globals** (i.e., global constants broadly
concerning various third-party modules and packages rather than one specific
third-party module or package).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.typing.datatyping import DictStrToFrozenSetStrs
from beartype._util.kind.maplike.utilmapfrozen import FrozenDict

# ....................{ DICTS                              }....................
BLACKLIST_MODULE_NAME_TO_TYPE_NAMES: DictStrToFrozenSetStrs = FrozenDict({
    # ....................{ ANTIPATTERN ~ decor-hostile    }....................
    # These third-party packages and modules widely employ the decorator-hostile
    # decorator antipattern throughout their codebases and are thus
    # runtime-hostile.

    # The object-oriented @fastmcp.FastMCP.tool decorator method destructively
    # transforms callable user-defined functions and methods into *UNCALLABLE*
    # FastMCP-specific instances of this type. Why, FastMCP!? WHY!?!!?? *sigh*
    'fastmcp.tools.tool': frozenset(('FunctionTool',)),

    # The object-oriented @langchain_core.runnables.chain decorator method
    # destructively transforms callable user-defined functions and methods into
    # *UNCALLABLE* LangChain-specific instances of this type. Why, LangChain!?!
    'langchain_core.runnables.base': frozenset(('RunnableLambda',)),
})
'''
Frozen dictionary mapping from the fully-qualified name of each problematic
third-party package and module to a frozen set of the unqualified basenames of
all **beartype-blacklisted types** defined by that package or module. These
types are well-known to be hostile to runtime type-checking in general and
:mod:`beartype` specifically, usually due to employing one or more of the
following antipatterns:

* The **decorator-hostile decorator antipattern,** a harmful design pattern
  unofficially promoted throughout the large language model (LLM) community.
  This antipattern abuses the standard PEP-compliant decorator paradigm (which
  supports decorator chaining by permitting arbitrary decorators to be applied
  to other decorators) by prohibiting decorator chaining. Many open-source LLM
  APIs, for example, define decorator-hostile decorators destructively transform
  callable user-defined functions and methods into uncallable instances of
  non-standard types usable *only* by those APIs. Due to being uncallable *and*
  non-standard, those instances then obstruct trivial wrapping by the
  :func:`beartype.beartype` decorator.
'''


BLACKLIST_TYPE_MRO_ROOT_MODULE_NAME_TO_TYPE_NAMES: DictStrToFrozenSetStrs = (
    FrozenDict({
        # ....................{ ANTIPATTERN ~ decor-hostile    }....................
        # These third-party packages and modules widely employ the decorator-hostile
        # decorator antipattern throughout their codebases and are thus
        # runtime-hostile.

        # The object-oriented @celery.Celery.task decorator method transforms
        # callable user-defined functions and methods into callable
        # Celery-specific instances of this type, known as Celery tasks. For
        # better or worse, Celery tasks masquerade as the user-defined callables
        # they wrap and thus are *ONLY* accessible as the root MRO item:
        #     # Define a trivial Celery task.
        #     >>> from celery import Celery
        #     >>> celery_server = Celery(broker='memory://')
        #     >>> @celery_server.task()
        #     >>> def muh_celery_task() -> None: pass
        #
        #     # Prove that Celery tasks lie about everything.
        #     >>> muh_celery_task.__module__
        #     celery.local  # <-- weird, but okay
        #     >>> muh_celery_task.__name__
        #     muh_celery_task  # <-- *LIAR*! you're actually a "Task" instance!!
        #     >>> muh_celery_task.__class__.__module__
        #     __main__  # <-- *LIAR*! you're actually a "Task" instance!!
        #     >>> muh_celery_task.__class__.__name__
        #     muh_celery_task  # <-- *LIAR*! you're actually a "Task" instance!!
        #     >>> muh_celery_task.__class__.__mro__
        #     (<class '__main__.muh_celery_task'>, <class
        #     'celery.app.task.Task'>, <class 'celery.app.task.Task'>, <class
        #     'object'>)  # <-- *FINALLY*. at last. the truth is revealed.
        'celery.app.task': frozenset(('Task',)),
    }))
'''
Frozen dictionary mapping from the fully-qualified name of each problematic
third-party package and module to a frozen set of the unqualified basenames of
all **beartype-blacklisted types** defined by that package or module such that
these high-level types masquerade as the low-level user-defined callables that
they wrap, typically by an even higher-level decorator wrapping callables with
those types.

These types hide themselves from public view and thus are *only* accessible as
the **root method-resolution order (MRO) item** (i.e., second-to-last item of
the ``__mro__`` dunder dictionary of these types, thus ignoring the ignorable
:class:`object` guaranteed to be the last item of all such dictionaries).

See Also
--------
:data:`.BLACKLIST_MODULE_NAME_TO_TYPE_NAMES`
    Further details.
'''

# ....................{ SETS                               }....................
#FIXME: Apply this blacklist to the following things:
#* Arbitrary callables to be decorated by @beartype, possibly. Consider defining
#  a new beartype._util.bear.utilbearfunc.is_func_thirdparty_blacklisted()
#  tester returning True *ONLY* if the passed callable has a "__module__" dunder
#  attribute whose value is a string residing in this frozenset.
BLACKLIST_PACKAGE_NAMES = frozenset((
    # ....................{ ANTIPATTERN ~ forward ref      }....................
    # These third-party packages and modules widely employ the forward reference
    # antipattern throughout their codebases and are thus runtime-hostile.

    # Pydantic employs the forward reference antipattern everywhere: e.g.,
    #     from __future__ import annotations as _annotations
    #     import typing
    #     ...
    #
    #     if typing.TYPE_CHECKING:
    #         ...
    #         from ..main import BaseModel  # <-- undefined at runtime
    #
    #     ...
    #     def wrapped_model_post_init(
    #         self: BaseModel, context: Any, /) -> None:  # <-- unresolvable at runtime
    #
    # See also this @beartype-specific issue on Pydantic:
    #     https://github.com/beartype/beartype/issues/444
    'pydantic',

    # "urllib3" employs the forward reference antipattern everywhere: e.g.,
    #     import typing
    #     ...
    #
    #     if typing.TYPE_CHECKING:
    #         from typing import Final  # <-- undefined at runtime
    #
    #     ...
    #
    #     _DEFAULT_TIMEOUT: Final[_TYPE_DEFAULT] = _TYPE_DEFAULT.token
    #
    # See also this @beartype-specific comment on "urllib3":
    #     https://github.com/beartype/beartype/issues/223#issuecomment-2525261497
    'urllib3',

    # xarray employs the forward reference antipattern everywhere: e.g.,
    #     from __future__ import annotations
    #     from typing import IO, TYPE_CHECKING, Any, Generic, Literal, cast, overload
    #     ...
    #
    #     if TYPE_CHECKING:
    #         ...
    #         from xarray.core.dataarray import DataArray  # <-- undefined at runtime
    #
    #     ...
    #
    #     class Dataset(
    #         DataWithCoords,
    #         DatasetAggregations,
    #         DatasetArithmetic,
    #         Mapping[Hashable, "DataArray"],  # <-- unresolvable at runtime
    #     ):
    #
    # See also this @beartype-specific issue on xarray:
    #     https://github.com/beartype/beartype/issues/456
    'xarray',
))
'''
Frozen set of the fully-qualified names of all **beartype-blacklisted
third-party packages** well-known to be hostile to runtime type-checking in
general and :mod:`beartype` specifically, usually due to employing one or more
of the following antipatterns:

* The **forward reference antipattern,** a `harmful design pattern officially
  promoted throughout "mypy" documentation <antipattern_>`__. This antipattern
  leverages both :pep:`563` *and* the :pep:`484`-compliant
  :obj:`typing.TYPE_CHECKING` global (both of which are well-known to be hostile
  to runtime type-checking) to conditionally define relative forward references
  visible *only* to pure static type-checkers like ``mypy`` and ``pyright``.
  These references are undefined at runtime and thus inaccessible to hybrid
  runtime-static type-checkers like :mod:`beartype` and :mod:`typeguard`.

  As an example, consider this hypothetical usage of the forward reference
  antipattern in a mock third-party package named ``"awful_package"``:

  .. code-block:: python

     from __future__ import annotations  # <-- pep 563
     from typing import TYPE_CHECKING    # <-- pep 484

     if TYPE_CHECKING:                         # <---- "False" at runtime
         from awful_package import AwfulClass  # <-- undefined at runtime

     # PEP 563 (i.e., "from __future__ import annotations") stringifies the
     # undefined "AwfulClass" class to the string "'AwfulClass'" at runtime.
     # Since the "AwfulClass" class is undefined, however, neither @beartype nor
     # any other runtime type-checker can resolve this relative forward
     # reference to the external "awful_package.AwfulClass" class it refers to.
     def awful_func(awful_arg: AwfulClass): ...

.. _antipattern:
   https://mypy.readthedocs.io/en/latest/runtime_troubles.html#import-cycles
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-agnostic type hint getter utilities** (i.e., callables
validating querying type hints supported by :mod:`beartype`, regardless of
whether those hints comply with PEP standards or not).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._util.cache.utilcachecall import callable_cached

# ....................{ TESTERS                            }....................
#FIXME: Call this getter everywhere we currently call "repr(hint*)", please.
@callable_cached
def get_hint_repr(hint: object) -> str:
    '''
    **Representation** (i.e., machine-readable string returned by the
    :func:`repr` builtin when passed this hint) of this PEP-agnostic type hint.

    This getter is memoized for efficiency. Indeed, since the implementation of
    this getter trivially reduces to just ``repr(hint)``, memoization is the
    entire reason for the existence of this getter.

    Motivation
    ----------
    **This getter should always be called to efficiently obtain the
    representation of any type hint.** The comparatively less efficient
    :func:`repr` builtin should *never* be called to do so.

    Parameters
    ----------
    hint : object
        Type hint to be represented.

    Returns
    ----------
    str
        Representation of this hint.
    '''

    # Return the machine-readable representation of this hint.
    return repr(hint)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **intraprocess import hook Python optimization integration tests**
(i.e., validating :mod:`beartype.claw` import hooks with respect to optimization
of the active Python interpreter by either the ``-O`` command-line option *or*
the ``${PYTHONOPTIMIZE}`` environment variable).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: Isolate each integration test defined below to its own subprocess.
# Why? Module imports. Since each integration test defined below tends to
# reimport the same (or, at least, similar) modules as previously run unit tests
# defined below, module imports and thus unit tests *MUST* be isolated to their
# own subprocesses to ensure these tests may be run in any arbitrary order.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
import pytest

# ....................{ TESTS                              }....................
@pytest.mark.run_in_subprocess
def test_claw_optimized() -> None:
    '''
    Test :mod:`beartype.claw` import hooks against a Python
    optimization-specific data subpackage in this test suite exercising these
    hooks against **Python optimization** (i.e., the ``-O`` command-line option
    *and* ``${PYTHONOPTIMIZE}`` environment variable).

    Note that what *should* be a single integration test is intentionally
    subdivided into two integration tests -- this and the sibling
    :func:`.test_claw_unoptimized` test. Although unifying both into a single
    integration test would certainly be preferable, doing so is complicated by
    the need to reload the Python optimization-specific data package to which
    :mod:`beartype.claw` import hooks are applied. Since Python lacks any
    official support for module reloading, these two integration tests are
    instead isolated to subprocesses -- trivially circumventing the need for
    module reloading in the first place.

    See Also
    --------
    :func:`.test_claw_unoptimized`
        Companion integration test.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.claw import beartype_package
    from pytest import MonkeyPatch

    # ....................{ PASS                           }....................
    # Inside the equivalent of the "monkeypatch" fixture...
    with MonkeyPatch.context() as monkeypatch:
        # Temporarily set the Python optimization state by setting the
        # "${PYTHONOPTIMIZE}" environment variable to a positive integer.
        monkeypatch.setenv('PYTHONOPTIMIZE', '1')

        # Subject a Python optimization-specific data package to a beartype
        # import hook, configured by the default beartype configuration.
        beartype_package('beartype_test.a90_func.data.claw.internal.optimized')

        # Import all submodules of the package hooked above, exercising that
        # these submodules are subject to that import hook *AND* satisfy all
        # integration tests implemented as assertion statements in these
        # submodules.
        from beartype_test.a90_func.data.claw.internal.optimized import (
            data_claw_optimized)


@pytest.mark.run_in_subprocess
def test_claw_unoptimized() -> None:
    '''
    Test :mod:`beartype.claw` import hooks against a Python
    optimization-specific data subpackage in this test suite exercising these
    hooks against a lack of **Python optimization** (i.e., the ``-O``
    command-line option *and* ``${PYTHONOPTIMIZE}`` environment variable).

    See Also
    --------
    :func:`.test_claw_optimized`
        Companion integration test.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.claw import beartype_package

    # ....................{ PASS                           }....................
    # Subject a Python optimization-specific data package to a beartype
    # import hook, configured by the default beartype configuration.
    #
    # Note that, since this test suite is *ALWAYS* run unoptimized and this
    # integration test inherits that state, this integration test is also
    # implicitly run unoptimized. So it goes in Python world. \o/
    beartype_package('beartype_test.a90_func.data.claw.internal.optimized')

    # Import all submodules of the package hooked above, exercising that
    # these submodules are subject to that import hook *AND* satisfy all
    # integration tests implemented as assertion statements in these
    # submodules.
    from beartype_test.a90_func.data.claw.internal.optimized import (
        data_claw_optimized)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **decorator position globals** (i.e., global constants defining the
initial user-configurable contents of the beforelist automating decorator
positioning for :mod:`beartype.claw` import hooks).

:mod:`beartype.claw` import hooks initialize user-configurable beforelists via
these globals of third-party decorators well-known to be **decorator-hostile**
(i.e., decorators hostile to other decorators by prematurely terminating
decorator chaining, such that *no* decorators may appear above these decorators
in any chain of one or more decorators).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Dict,
    Optional,
)
from beartype._conf.decorplace.confplacetrie import (
    BeartypeDecorPlacePackagesTrie,
    BeartypeDecorPlacePackageTrie,
    BeartypeDecorPlaceTypeTrie,
)

# ....................{ HINTS                              }....................
#FIXME: Actually define these hints as proper type aliases *AFTER* we drop
#support for Python 3.11, please: e.g.,
#   type _ClawBeforelist = Dict[str, Union[FrozenSet[str], _ClawBeforelist]]

BeartypeDecorPlaceSubtrie = Optional[Dict[str, 'BeartypeDecorPlaceSubtrie']]
'''
PEP-compliant recursive alias matching a **beforelist subtrie** (i.e.,
corresponding value of some key-value pair signifying a child node of a parent
beforelist (sub)trie), constrained to be either:

* If this is a leaf child node terminating a branch of that (sub)trie,
  :data:`None`.
* If this is a stem child node perpetuating a branch of that (sub)trie, yet
  another such recursively nested dictionary mapping from the unqualified
  basenames of problematic third-party attributes imported into a scope of the
  currently visited module to yet another (sub)trie child node.
'''


BeartypeDecorPlaceTrie = Dict[str, Dict[str, BeartypeDecorPlaceSubtrie]]
'''
PEP-compliant recursive alias matching a **beforelist trie** (i.e., recursive
tree structure whose nodes are the unqualified basenames of problematic
third-party attributes imported into a scope of the currently visited module,
defined as a frozen dictionary mapping from strings to either yet another such
recursively nested frozen dictionary *or* :data:`None` signifying a terminal
leaf node).

Note that the root trie is guaranteed to map from strings to *only* nested
frozen dictionaries (rather than to both nested frozen dictionaries and
:data:`None`). Consequently, this hint intentionally differentiates between
matching the root and non-root nesting levels of this trie.
'''

# ....................{ TRIES                              }....................
#FIXME: *BEFORE* openly publishing this data structure to the world, let's do
#something about the hideously ambiguous "None" references we're stuffing into
#this data structure. Specifically:
#* Define a new "BeartypeDecorPlaceNode" class and associated
#  "BeartypeDecorPlaceDecoratorHostileNode" global singleton: e.g.,
#      class BeartypeDecorPlaceNode(object):
#          pass
#
#      BeartypeDecorPlaceDecoratorHostileNode = BeartypeDecorPlaceNode()
#* Globally replace *ALL* ambiguous "None" references both here and in the
#  associated "clawastimport" submodule with unambiguous
#  "BeartypeDecorPlaceDecoratorHostileNode" references instead.
#
#Why are "None" references ambiguous? Because we'd like to eventually stuff a
#wide variety of metadata into this data structure. For example, third-party
#*DECORATOR-DISABLING DECORATORS*. "But what are decorator-disabling
#decorators!?", you might now be cogitating. As the term suggests, they're
#decorators whose existence in a chain of one or more decorators signals to
#other decorators that those decorators should *NOT* be applied in the first
#place. A great real-world example of a decorator-disabling decorator is the
#@jaxtyping.jaxtyped decorator, which already internally applies a runtime
#type-checking decorator like @beartype; ergo, @beartype should *NOT* be
#erroneously re-applied to callables and types decorated by @jaxtyping.jaxtyped.
#That decorator effectively disables other decorators.
#
#Unambiguous node singletons give us the future flexibility we need to
#eventually support features like this. Pump that fist, bear bros! \o/
DECOR_HOSTILE_ATTR_NAME_TRIE: BeartypeDecorPlaceTrie = (
    BeartypeDecorPlacePackagesTrie({
        # ....................{ FUNCTIONS                  }....................
        # Third-party decorator-hostile decorator *FUNCTIONS* directly defined
        # by functional (i.e., *NOT* object-oriented) APIs.

        # The third-party @chain decorator function of the
        # "langchain_core.runnables" package of the LangChain API. See also:
        #     https://github.com/beartype/beartype/issues/541
        'langchain_core': BeartypeDecorPlacePackageTrie(
            {'runnables': BeartypeDecorPlacePackageTrie({'chain': None})}),

        # ....................{ METHODS                    }....................
        # Third-party decorator-hostile decorator *METHODS* directly defined by
        # types directly defined by object-oriented (OO) APIs.

        # The third-party @task decorator method of the "celery.Celery" type of
        # the Celery API. See also:
        #     https://github.com/beartype/beartype/issues/500
        'celery': BeartypeDecorPlacePackageTrie({
            'Celery': BeartypeDecorPlaceTypeTrie({'task': None})}),

        # The third-party @tool decorator function of the "fastmcp.FastMCP" type
        # of the FastMCP API. See also:
        #     https://github.com/beartype/beartype/issues/540
        'fastmcp': BeartypeDecorPlacePackageTrie({
            'FastMCP': BeartypeDecorPlaceTypeTrie({'tool': None})}),
}))
'''
**Decorator-hostile decorator attribute name trie** (i.e., frozen dictionary
mapping from the unqualified basename of each third-party (sub)package and
(sub)module transitively defining one or more decorator-hostile decorators to
either yet another such recursively nested frozen dictionary *or* :data`None`,
in which case the corresponding key is the unqualified basename of a
decorator-hostile decorator directly defined by that (sub)package, (sub)module,
type, or instance).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype import hook configuration subpackage initialization submodule**
(i.e., data module mimicking real-world usage of various :func:`beartype.claw`
import hooks configured by different beartype configurations).
'''

# ....................{ IMPORTS                            }....................
from beartype import (
    BeartypeConf,
    BeartypeDecorPlace,
)
from beartype.claw import beartype_package

# Subject this single module to a beartype import hook configured to inject the
# @beartype decorator first before (i.e., below) all other class decorators.
beartype_package(
    package_name=(
        'beartype_test.a00_unit.data.claw.intraprocess.hookable_package.conf.'
        'data_claw_conf_decorplace_funcs_first'
    ),
    conf=BeartypeConf(
        claw_decor_place_func=BeartypeDecorPlace.FIRST),
)
from beartype_test.a00_unit.data.claw.intraprocess.hookable_package.conf import (
    data_claw_conf_decorplace_funcs_first)

# Subject this single module to a beartype import hook configured to inject the
# @beartype decorator first before (i.e., below) all other class decorators.
beartype_package(
    package_name=(
        'beartype_test.a00_unit.data.claw.intraprocess.hookable_package.conf.'
        'data_claw_conf_decorplace_types_first'
    ),
    conf=BeartypeConf(
        claw_decor_place_type=BeartypeDecorPlace.FIRST),
)
from beartype_test.a00_unit.data.claw.intraprocess.hookable_package.conf import (
    data_claw_conf_decorplace_types_first)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype validator text utility unit tests.**

This submodule unit tests the private
:mod:`beartype.vale._util._valeutiltext` submodule.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ class : subscriptedis     }....................
def test_api_vale_format_diagnosis_line() -> None:
    '''
    Test successful usage of the private
    :func:`beartype.vale._util._valeutiltext.format_diagnosis_line` formatter.
    '''

    # Defer test-specific imports.
    from beartype.vale._util._valeutiltext import format_diagnosis_line

    # Assert this formatter accepts a true boolean value.
    visit_the_soul_in_sleep = format_diagnosis_line(
        validator_repr='Some say that gleams of a remoter world',
        indent_level_outer='    ',
        indent_level_inner='',
        is_obj_valid=True,
    )
    assert 'Some say that gleams of a remoter world' in visit_the_soul_in_sleep
    assert 'True' in visit_the_soul_in_sleep

    # Assert this formatter accepts a true boolean value.
    that_death_is_slumber = format_diagnosis_line(
        validator_repr='And that its shapes the busy thoughts outnumber',
        indent_level_outer='    ',
        indent_level_inner='',
        is_obj_valid=False,
    )
    assert 'And that its shapes the busy thoughts outnumber' in (
        that_death_is_slumber)
    assert 'False' in that_death_is_slumber

    # Assert this formatter accepts *NO* boolean value.
    i_look_on_high = format_diagnosis_line(
        validator_repr='Of those who wake and live.',
        indent_level_outer='    ',
        indent_level_inner='',
    )
    assert 'Of those who wake and live.' in i_look_on_high
    assert 'True'  not in i_look_on_high
    assert 'False' not in i_look_on_high



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype type-checking function code utility factories** (i.e., low-level
callables dynamically generating pure-Python code snippets type-checking
arbitrary objects passed to arbitrary callables against PEP-compliant type hints
passed to those same callables).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.code.datacodename import (
    ARG_NAME_GETRANDBITS,
)
from beartype._check.signature._sigsnip import (
    CODE_SIGNATURE_SCOPE_ARG_format,
    CODE_INIT_RANDOM_INT,
)
from beartype._conf.confmain import BeartypeConf
from beartype._data.typing.datatyping import (
    LexicalScope,
)
from beartype._util.text.utiltextrepr import represent_object

# ....................{ MAKERS ~ signature                 }....................
#FIXME: Unit test us up, please.
def make_func_signature(
    # Mandatory parameters.
    func_name: str,
    func_scope: LexicalScope,
    code_signature_format: str,
    conf: BeartypeConf,

    # Optional parameters.
    code_signature_prefix: str = '',
) -> str:
    '''
    **Type-checking signature factory** (i.e., low-level function dynamically
    generating and returning the **signature** (i.e., callable declaration
    prefixing the body of that callable) of a callable type-checking arbitrary
    objects against arbitrary type hints, described by the passed parameters.

    Parameters
    ----------
    func_name : str
        Unqualified basename of the callable declared by this signature.
    func_scope : LexicalScope
        **Local scope** (i.e., dictionary mapping from the name to value of
        each hidden parameter declared in this signature) of that callable,
        where a "hidden parameter" is a parameter whose name is prefixed by
        ``"__beartype_"`` and whose value is that of an external attribute
        internally referenced in the body of that callable.
    code_signature_format : str
        Code snippet declaring the unformatted signature of that callable, which
        this factory then formats by replacing these format variables in this
        code snippet:

        * ``{func_name}``, replaced by the value of the ``func_name`` parameter.
        * ``{code_signature_prefix}``, replaced by the value of the
          ``code_signature_prefix`` parameter.
        * ``{code_signature_scope_args}``, replaced by the declaration of all
          hidden parameters in the passed ``func_scope`` parameter.
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object).
    code_signature_prefix : str, optional
        Code snippet prefixing this signature, typically either:

        * If a synchronous callables, the empty string.
        * For asynchronous callables (e.g., asynchronous generators,
          coroutines), the space-suffixed keyword ``"async "``.

        Defaults to the empty string and thus synchronous behaviour.

    Yields
    ------
    str
        Signature of this callable.
    '''
    assert isinstance(func_name, str), f'{repr(func_name)} not string.'
    assert isinstance(func_scope, dict), f'{repr(func_scope)} not dictionary.'
    assert isinstance(conf, BeartypeConf), f'{repr(conf)} not configuration.'
    assert isinstance(code_signature_format, str), (
        f'{repr(code_signature_format)} not string.')
    assert isinstance(code_signature_prefix, str), (
        f'{repr(code_signature_prefix)} not string.')

    # Python code snippet declaring all optional private beartype-specific
    # parameters directly derived from the local scope established by the above
    # calls to the _code_check_args() and _code_check_return() functions.
    code_signature_scope_args = ''

    # For the name and value of each such parameter...
    for arg_name, arg_value in func_scope.items():
        # Machine-readable representation of this parameter's initial value,
        # stripped of newline and truncated to a (hopefully) sensible length.
        # Since the represent_object() function called below to sanitize this
        # value is incredibly slow, this representation is conditionally
        # appended as a human-readable comment to the declaration of this
        # parameter below *ONLY* if the caller explicitly requested debugging.
        arg_comment = (
            f' # is {represent_object(arg_value)}'
            if conf.is_debug else
            ''
        )

        # Compose the declaration of this parameter in the signature of this
        # wrapper from...
        code_signature_scope_args += CODE_SIGNATURE_SCOPE_ARG_format(
            arg_name=arg_name, arg_comment=arg_comment)

    #FIXME: *YIKES.* We need to pass a unique tester function signature here
    #resembling:
    #    def {{func_name}}(obj: object) -> bool:
    #To do so sanely, let's generalize this factory to accept an additional
    #mandatory "func_signature" parameter, please. We'll need to note in the
    #docstring exactly what format variables that parameter is expected to
    #contain, of course.

    # Python code snippet declaring the signature of this wrapper.
    code_signature = code_signature_format.format(
        func_name=func_name,
        code_signature_prefix=code_signature_prefix,
        code_signature_scope_args=code_signature_scope_args,
    )

    # Python code snippet of preliminary statements (e.g., local variable
    # assignments) if any *AFTER* generating snippets type-checking parameters
    # and returns (which modifies dataclass variables tested below).
    code_body_init = (
        # If the body of this wrapper requires a pseudo-random integer, append
        # code generating and localizing such an integer to this signature.
        CODE_INIT_RANDOM_INT
        if ARG_NAME_GETRANDBITS in func_scope else
        # Else, this body requires *NO* such integer. In this case, preserve
        # this signature as is.
        ''
    )

    # Return this signature suffixed by zero or more preliminary statements.
    return f'{code_signature}{code_body_init}'



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :mod:`contextlib` utility unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.api.standard.utilcontextlib` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ getter                     }....................
def test_get_func_contextlib_contextmanager_or_none() -> None:
    '''
    Test the
    :func:`beartype._util.api.standard.utilcontextlib.get_func_contextlib_contextmanager_or_none`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.api.standard.utilcontextlib import (
        get_func_contextlib_contextmanager_or_none)
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11
    from beartype_test.a00_unit.data.data_type import (
        async_context_manager_factory,
        context_manager_factory,
        function,
    )
    from beartype_test.a00_unit.data.func.data_decor import (
        decorator_isomorphic,
        decorator_nonisomorphic,
    )
    from contextlib import (
        asynccontextmanager,
        contextmanager,
    )

    # ....................{ PASS                           }....................
    # Assert this getter either accepts or rejects isomorphic closures created
    # and returned by @contextlib decorators conditionally depending on whether
    # the active Python interpreter targets Python >= 3.11 or not.
    if IS_PYTHON_AT_LEAST_3_11:
        assert get_func_contextlib_contextmanager_or_none(
            async_context_manager_factory) is asynccontextmanager
        assert get_func_contextlib_contextmanager_or_none(
            context_manager_factory) is contextmanager
    else:
        assert get_func_contextlib_contextmanager_or_none(
            async_context_manager_factory) is None
        assert get_func_contextlib_contextmanager_or_none(
            context_manager_factory) is None

    # ....................{ FAIL                           }....................
    # Assert this getter rejects isomorphic closures *NOT* created and returned
    # by the @contextlib.contextmanager decorator.
    assert get_func_contextlib_contextmanager_or_none(
        decorator_isomorphic(function)) is None

    # Assert this getter rejects non-isomorphic closures.
    assert get_func_contextlib_contextmanager_or_none(
        decorator_nonisomorphic(function)) is None

    # Assert this getter rejects non-closure callables.
    assert get_func_contextlib_contextmanager_or_none(function) is None

    # Assert this getter rejects non-callables.
    assert get_func_contextlib_contextmanager_or_none(
        'A lovely youth,—no mourning maiden decked') is None



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Sphinx utilities** (i.e., low-level callables handling the
third-party :mod:`sphinx` package as an optional runtime dependency of this
project).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To prevent this project from accidentally requiring third-party
# packages as mandatory runtime dependencies, avoid importing from *ANY* such
# package via a module-scoped import. These imports should be isolated to the
# bodies of callables declared below.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype._util.func.utilfuncframe import iter_frames
from sys import modules as module_imported_names

# ....................{ TESTERS                            }....................
def is_sphinx_autodocing() -> bool:
    '''
    :data:`True` only if Sphinx is currently **autogenerating documentation**
    (i.e., if this function has been called from a Python call stack invoked by
    the ``autodoc`` extension bundled with the optional third-party build-time
    :mod:`sphinx` package).
    '''

    # If the "autodoc" extension has *NOT* been imported, Sphinx by definition
    # *CANNOT* be autogenerating documentation. In this case, return false.
    #
    # Note this technically constitutes an optional (albeit pragmatically
    # critical) optimization. This test is O(1) with negligible constants,
    # whereas the additional test below is O(n) with non-negligible constants.
    # Ergo, this efficient test short-circuits the inefficient test below.
    if _SPHINX_AUTODOC_SUBPACKAGE_NAME not in module_imported_names:
        return False
    # Else, the "autodoc" extension has been imported. Since this does *NOT*
    # conclusively imply that Sphinx is currently autogenerating documentation,
    # further testing is required to avoid returning false positives (and thus
    # erroneously reducing @beartype to a noop, which would be horrifying).
    #
    # Specifically, we iteratively search up the call stack for a stack frame
    # originating from the "autodoc" extension. If we find such a stack frame,
    # Sphinx is currently autogenerating documentation; else, Sphinx is not.

    #FIXME: Refactor this to leverage a genuinely valid working solution
    #hopefully provided out-of-the-box by some hypothetical new bleeding-edge
    #version of Sphinx *AFTER* they resolve our feature request for this:
    #    https://github.com/sphinx-doc/sphinx/issues/9805

    # For each stack frame on the call stack, ignoring the stack frame
    # encapsulating the call to this tester...
    for frame in iter_frames(func_stack_frames_ignore=1):
        # Fully-qualified name of this scope's module if this scope defines
        # this name *OR* "None" otherwise.
        frame_module_name = frame.f_globals.get('__name__')
        # print(f'Visiting frame (module: "{func_frame_module_name}")...')

        # If this scope's module is the "autodoc" extension, Sphinx is
        # currently autogenerating documentation. In this case, return true.
        if (
            frame_module_name and
            frame_module_name.startswith(_SPHINX_AUTODOC_SUBPACKAGE_NAME)
        ):
            return True
        # Else, this scope's module is *NOT* the "autodoc" extension.

    # Else, *NO* scope's module is the "autodoc" extension. Return false.
    return False

# ....................{ PRIVATE ~ magic                    }....................
_SPHINX_AUTODOC_SUBPACKAGE_NAME = 'sphinx.ext.autodoc'
'''
Fully-qualified name of the subpackage providing the ``autodoc`` extension
bundled with Sphinx.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`585`-compliant **generic reducers** (i.e.,
low-level callables converting higher-level subscripted and unsubscripted
generics to lower-level type hints more readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Generic,
    Optional,
)
from beartype._data.typing.datatypingport import Hint
from beartype._check.metadata.hint.hintsane import (
    HINT_SANE_IGNORABLE,
    HintOrSane,
    HintSane,
)
from beartype._util.hint.pep.proposal.pep544 import is_hint_pep484_generic_io
from beartype._util.hint.pep.utilpepget import get_hint_pep_origin_or_none

# ....................{ REDUCERS                           }....................
def reduce_hint_pep484585_generic_subbed(
    hint: Hint,
    hint_parent_sane: Optional[HintSane],
    exception_prefix: str,
    **kwargs,
) -> HintOrSane:
    '''
    Reduce the passed :pep:`484`- or :pep:`585`-compliant **subscripted
    generic** (i.e., object subscripted by one or more child type hints
    originating from a type originally subclassing at least one subscripted
    :pep:`484`- or :pep:`585`-compliant pseudo-superclass) to a more suitable
    type hint better supported by :mod:`beartype` if necessary.

    This reducer ignores *all* parametrizations of the :class:`typing.Generic`
    abstract base class (ABC) by one or more type variables. As the name
    implies, this ABC is generic and thus fails to impose any meaningful
    constraints. Since a type variable in and of itself also fails to impose any
    meaningful constraints, these parametrizations are safely ignorable in all
    possible contexts: e.g.,

    .. code-block:: python

       from typing import Generic, TypeVar
       T = TypeVar('T')
       def noop(param_hint_ignorable: Generic[T]) -> T: pass

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : Hint
        Subscripted generic to be reduced.
    hint_parent_sane : Optional[HintSane]
        Either:

        * If the passed hint is a **root** (i.e., top-most parent hint of a tree
          of child hints), :data:`None`.
        * Else, the passed hint is a **child** of some parent hint. In this
          case, the **sanified parent type hint metadata** (i.e., immutable and
          thus hashable object encapsulating *all* metadata previously returned
          by :mod:`beartype._check.convert.convmain` sanifiers after sanitizing
          the possibly PEP-noncompliant parent hint of this child hint into a
          fully PEP-compliant parent hint).
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    All remaining passed keyword parameters are silently ignored.

    Returns
    -------
    HintOrSane
        Either:

        * If the unsubscripted hint (e.g., :class:`typing.Generic`) originating
          this subscripted hint (e.g., ``typing.Generic[S, T]``) is
          unparametrized by type variables, that unsubscripted hint as is.
        * Else, that unsubscripted hint is parametrized by one or more type
          variables. In this case, the **sanified type hint metadata** (i.e.,
          :class:`.HintSane` object) describing this reduction.
    '''

    # Avoid circular import dependencies.
    from beartype._check.convert._reduce._pep.redpep484612646 import (
        reduce_hint_pep484612646_subbed_typeargs_to_hints)

    # If this subscripted generic is the "typing.Generic" superclass directly
    # parametrized by one or more type variables (e.g., "typing.Generic[T]"),
    # this generic is ignorable. In this case, reduce this ignorable generic to
    # the ignorable singleton.
    #
    # Note that we intentionally avoid calling the
    # get_hint_pep_origin_type_isinstanceable_or_none() function here, which has
    # been intentionally designed to exclude PEP-compliant type hints
    # originating from "typing" type origins for stability reasons.
    if get_hint_pep_origin_or_none(hint) is Generic:
        # print(f'Testing generic hint {repr(hint)} deep ignorability... True')
        return HINT_SANE_IGNORABLE
    # Else, this subscripted generic is *NOT* the "typing.Generic" superclass
    # directly parametrized by one or more type variables and thus *NOT* an
    # ignorable non-protocol.
    #
    # Note that this condition being false is *NOT* sufficient to declare this
    # hint to be unignorable. Notably, the origin type originating both
    # ignorable and unignorable protocols is "Protocol" rather than "Generic".
    # Ergo, this generic could still be an ignorable protocol.

    # Useful PEP 544-compliant unsubscripted protocol possibly reduced from this
    # useless PEP 484- or 585-compliant subscripted IO generic if this hint is a
    # subscripted IO generic *OR* this hint as is otherwise.
    #
    # Note that we reduce this subscripted IO generic *BEFORE* stripping all
    # child hints subscripting this IO generic, as this reducers requires these
    # child hints to correctly reduce this IO generic.
    hint_reduced = _reduce_hint_pep484585_generic_io(hint, exception_prefix)

    # If this hint was reduced to an unsubscripted generic from this subscripted
    # IO generic, return this reduced hint.
    if hint_reduced is not hint:
        return hint_reduced
    # Else, this hint was *NOT* reduced to an unsubscripted generic from this
    # subscripted IO generic (i.e., "hint_reduced is hint").

    # Reduce this subscripted generic to:
    # * The semantically useful unsubscripted generic originating this
    #   semantically useless subscripted generic.
    # * The type variable lookup table mapping all type variables parametrizing
    #   this unsubscripted generic to all non-type variable hints subscripting
    #   this subscripted generic.
    # print(f'[reduce_hint_pep484585_generic_subbed] Reducing subscripted generic {repr(hint)}...')
    hint_reduced = reduce_hint_pep484612646_subbed_typeargs_to_hints(
        hint=hint,
        hint_parent_sane=hint_parent_sane,
        exception_prefix=exception_prefix,
    )
    # print(f'[reduce_hint_pep484585_generic_subbed] ...to unsubscripted generic {repr(hint_reduced)}.')

    # Return this reduced hint.
    return hint_reduced


def reduce_hint_pep484585_generic_unsubbed(
    hint: Hint, exception_prefix: str) -> HintOrSane:
    '''
    Reduce the passed :pep:`484`- or :pep:`585`-compliant **unsubscripted
    generic** (i.e., type originally subclassing at least one unsubscripted
    :pep:`484`- or :pep:`585`-compliant pseudo-superclass) to a more suitable
    type hint better supported by :mod:`beartype` if necessary.

    This reducer is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : Hint
        Subscripted generic to be reduced.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages.

    Returns
    -------
    HintOrSane
        Either:

        * If this unsubscripted generic is ignorable, the
          :data:`.HINT_SANE_IGNORABLE` singleton.
        * Else, this unsubscripted generic possibly reduced to a more suitable
          hint.
    '''

    # If this unsubscripted generic is the "typing.Generic" superclass, this
    # generic is ignorable. In this case, reduce this ignorable generic to the
    # ignorable singleton.
    #
    # Note that we intentionally avoid calling the
    # get_hint_pep_origin_type_isinstanceable_or_none() function here, which has
    # been intentionally designed to exclude PEP-compliant type hints
    # originating from "typing" type origins for stability reasons.
    if hint is Generic:
        # print(f'Testing generic hint {repr(hint)} deep ignorability... True')
        return HINT_SANE_IGNORABLE
    # Else, this unsubscripted generic is *NOT* the "typing.Generic" superclass
    # and thus *NOT* an ignorable non-protocol.
    #
    # Note that this condition being false is *NOT* sufficient to declare this
    # hint to be unignorable. Notably, the origin type originating both
    # ignorable and unignorable protocols is "Protocol" rather than "Generic".
    # Ergo, this generic could still be an ignorable protocol.

    # Hint possibly reduced from this useless unsubscripted IO generic if this
    # hint is an unsubscripted IO generic *OR* this hint as is otherwise.
    hint_reduced = _reduce_hint_pep484585_generic_io(hint, exception_prefix)

    # Return this possibly reduced hint.
    return hint_reduced

# ....................{ PRIVATE ~ reducers                 }....................
def _reduce_hint_pep484585_generic_io(
    hint: Hint, exception_prefix: str) -> Hint:
    '''
    Reduce the passed :pep:`484`- or :pep:`585`-compliant **standard IO
    generic** (i.e., standard :obj:`typing.IO` generic (in either subscripted or
    unsubscripted forms) *or* the standard :obj:`BinaryIO` or :obj:`TextIO`
    unsubscripted generics) to a beartype-specific :pep:`544`-compliant protocol
    implementing this generic if this generic is a standard IO generic *or*
    silently ignore this generic otherwise.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : Hint
        Standard IO generic to be reduced.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages.

    All remaining passed keyword parameters are silently ignored.

    Returns
    -------
    Hint
        This subscripted generic possibly reduced to a more suitable hint.
    '''

    # Avoid circular import dependencies.
    from beartype._check.convert._reduce._pep.redpep544 import (
        reduce_hint_pep484_generic_io_to_pep544_protocol)

    # If this hint is a PEP 484-compliant IO generic base class, reduce this
    # functionally useless hint to the corresponding functionally useful
    # beartype-specific PEP 544-compliant protocol implementing this hint.
    if is_hint_pep484_generic_io(hint):
        # print(f'Reducing IO generic {repr(hint)}...')
        hint = reduce_hint_pep484_generic_io_to_pep544_protocol(
            hint, exception_prefix)
        # print(f'...{repr(hint)}.')
    # Else, this hint is *NOT* a PEP 484-compliant IO generic base class.
    # Preserve this hint as is.

    # Return this possibly reduced hint.
    return hint



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **weak reference** (i.e., references to objects explicitly
allowing those objects to be garbage-collected at *any* time) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilPythonWeakrefException
from beartype.typing import (
    Tuple,
)
from weakref import ref as weakref_ref

# ....................{ GETTERS                            }....................
def make_obj_weakref_and_repr(obj: object) -> Tuple[object, str]:
    '''
    2-tuple ``(weakref, repr)`` weakly referring to the passed object.

    Parameters
    ----------
    obj : object
        Arbitrary object to be weakly referred to.

    Returns
    ----------
    Tuple[object, str]
        2-tuple ``(weakref, repr)`` weakly referring to this object such that:

        * ``weakref`` is either:

          * If this object supports weak references, a **weak reference** (i.e.,
            :class:`weakref.ref` instance) to this object.
          * If this object prohibits weak references (e.g., due to being a
            common C-based variable-sized container like a tuple or string),
            ``None``.

        * ``repr`` is the machine-readable representation of this object,
          truncated to ~10KB to minimize space consumption in the worst case of
          an obscenely large object.
    '''

    # Avoid circular import dependencies.
    from beartype._util.text.utiltextrepr import represent_object

    # Weak reference to this object if this object supports weak references *OR*
    # "None" otherwise (e.g., if this object is a variable-sized container).
    obj_weakref = None

    # Machine-readable representation of this object truncated to minimize space
    # consumption for the worst case of an obscenely large object.
    obj_repr = represent_object(
        obj=obj,
        # Store at most 1KB of the full representation, which should
        # certainly suffice for most use cases. Note that the
        # default of 96B is far too small to be useful here.
        max_len=1000,
    )

    # If this object is "None", substitute "None" for this non-"None"
    # placeholder. Since the "weakref.ref" class ambiguously returns "None" when
    # this object has already been garbage-collected, this placeholder enables
    # subsequent calls to the get_obj_weakref_or_repr() getter to disambiguate
    # between these two common edge cases.
    if obj is None:
        obj_weakref = _WEAKREF_NONE
    # Else, this object is *NOT* "None". In this case...
    else:
        # Attempt to classify a weak reference to this object for safety.
        try:
            obj_weakref = weakref_ref(obj)
        # If doing so raises a "TypeError", this object *CANNOT* be weakly
        # referred to. Sadly, builtin variable-sized C-based types (e.g.,
        # "dict", "int", "list", "tuple") *CANNOT* be weakly referred to. This
        # constraint is officially documented by the "weakref" module:
        #     Several built-in types such as list and dict do not directly
        #     support weak references but can add support through subclassing.
        #     CPython implementation detail: Other built-in types such as tuple
        #     and int do not support weak references even when subclassed.
        #
        # Since this edge case is common, permitting this exception to unwind
        # the call stack is unacceptable; likewise, even coercing this exception
        # into non-fatal warnings would generic excessive warning spam and is
        # thus also unacceptable. The only sane solution remaining is to
        # silently store the machine-readable representation of this object and
        # return that rather than this object from the "object" property.
        except TypeError:
            pass

    return obj_weakref, obj_repr



def get_weakref_obj_or_repr(obj_weakref: object, obj_repr: str) -> object:
    '''
    Object weakly referred to by the passed object if this object is indeed a
    weak reference to another existing object *or* the passed machine-readable
    representation otherwise (i.e., if this object is either ``None`` *or* is a
    weak reference to a dead garbage-collected object).

    This function is typically passed the pair of objects returned by a prior
    call to the companion :func:`make_obj_weakref_and_repr` function.

    Parameters
    ----------
    obj_weakref : object
        Either:

        * If the **referent** (i.e., target object being weakly referred to) is
          the ``None`` singleton, the :data:`_WEAKREF_NONE` placeholder.
        * Else if the referent supports weak references, a **weak reference**
          (i.e., :class:`weakref.ref` instance) to that object.
        * Else, ``None``.
    obj_repr : str
        Machine-readable representation of that object, typically truncated to
        some number of characters to avoid worst-case space consumption.

    Returns
    ----------
    object
        Either:

        * If this weak reference is the :data:`_WEAKREF_NONE` placeholder, the
          ``None`` singleton.
        * Else if this referent support weak references, either:

          * If this referent is still alive (i.e., has yet to be
            garbage-collected), this referent.
          * Else, this referent is now dead (i.e., has already been
            garbage-collected). In this case, the passed representation.

        * Else, this referent does *not* support weak references (i.e., this
          weak reference is ``None``). In this case, the passed representation.

    Raises
    ----------
    _BeartypeUtilPythonWeakrefException
        If ``obj_weakref`` is invalid: i.e., neither ``None``,
        :data:`_WEAKREF_NONE`, nor a weak reference.
    '''
    assert isinstance(obj_repr, str), f'{repr(obj_repr)} not string.'

    # If this weak reference is "None", the prior call to
    # make_obj_weakref_and_repr() was passed an object that could *NOT* be
    # weakly referred to (e.g., C-based container). In this case, fallback to
    # the machine-readable representation of that object.
    if obj_weakref is None:
        return obj_repr
    # Else, this weak reference is *NOT* "None".
    #
    # If this weak reference is "_WEAKREF_NONE", the prior call to
    # make_obj_weakref_and_repr() was passed the "None" singleton. In this case,
    # substitute this placeholder for "None". See that factory.
    elif obj_weakref is _WEAKREF_NONE:
        return None
    # Else, this weak reference is *NOT* that placeholder.
    #
    # If this weak reference is *NOT* a weak reference, raise an exception.
    elif not isinstance(obj_weakref, weakref_ref):
        raise _BeartypeUtilPythonWeakrefException(
            f'Weak reference {repr(obj_weakref)} invalid '
            f'(i.e., neither weak reference, "None", nor "_WEAKREF_NONE").'
        )
    # Else, this weak reference is a weak reference.

    # Object weakly referred to by this weak reference if this object is alive
    # *OR* "None" otherwise (i.e., if this object was garbage-collected).
    obj = obj_weakref()

    # Return either...
    return (
        # If this object is still alive, this object;
        obj if obj is not None else
        # Else, this object is now dead. In this case, the machine-readable
        # representation of this object instead.
        obj_repr
    )

# ....................{ PROPERTIES ~ constants             }....................
_WEAKREF_NONE = object()
'''
Singleton substitute for the ``None`` singleton, enabling
:class:`BeartypeCallHintViolation` exceptions to differentiate between weak
references to ``None`` and weak references whose referents are already dead
(i.e., have already been garbage-collected).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator PEP-noncompliant Pandera type hint unit tests.**

This submodule unit tests the :func:`beartype.beartype` decorator with respect
to **PEP-noncompliant Pandera type hints** (i.e., :mod:`pandera.typing`-specific
annotations *not* compliant with annotation-centric PEPs).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_unless_package
from beartype_test._util.mark.pytmark import ignore_warnings

# ....................{ TESTS ~ pandas                     }....................
@skip_unless_package('pandera')
@skip_unless_package('pandas')
def test_pandera_pandas() -> None:
    '''
    Integration test validating that the :func:`beartype.beartype` decorator
    integrates cleanly with the third-party :class:`pandera.check_types`
    decorator with respect to Pandas dataframes.

    Specifically, this test validates that the :func:`beartype.beartype`
    decorator conditionally ignores *all* Pandera type hints.

    Motivation
    ----------
    Pandera type hints violate Python typing standards and thus require use of
    Pandera-specific functionality to type-check. This includes the non-standard
    :func:`pandera.check_types` decorator, which performs ad-hoc
    PEP-noncompliant runtime type-checking of ad-hoc PEP-noncompliant Pandera
    type hints annotating the decorated callable. Since type-checking Pandera
    type hints *de facto* requires usage of this decorator, the PEP-compliant
    :func:`beartype.beartype` decorator *cannot* natively type-check Pandera
    type hints. Instead, :func:`beartype.beartype` can only:

    * Detect that Pandera type hints are type-checked by a prior application of
      :func:`pandera.check_types` on the decorated callable.
    * If so, ignore *all* Pandera type hints.
    * Else, raise an exception.

    Whereas this test suite automates testing of PEP-compliant type hints via
    the :mod:`beartype_test.a00_unit.data` subpackage, Pandera type hints are
    fundamentally non-standard and thus *cannot* be automated in this standard
    manner. These hints can *only* be tested with a non-standard workflow
    implemented by this unit test.

    See Also
    --------
    https://pandera.readthedocs.io/en/stable/dataframe_models.html
        Official introduction to the :func:`pandera.check_types` decorator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports. Note that Pandas requires NumPy, which is
    # thus guaranteed to be safely importable here.
    from beartype import beartype
    from beartype.roar import (
        BeartypeCallHintParamViolation,
        BeartypeCallHintReturnViolation,
    )
    from pandas import (
        DataFrame,
        to_datetime,
    )
    from pandera.dtypes import (
        Int64,
        String,
        Timestamp,
    )
    from pandera.errors import SchemaError
    from pandera.pandas import (
        DataFrameModel as PandasDataFrameModel,
        check_types as pandas_check_types,
    )
    from pandera.typing import (
        DataFrame as PanderaDataFrame,
        Series,
    )
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Pandas data frame containing arbitrary sample data satisfying the
    # "PanderaModel" defined below.
    pandas_dataframe_good = DataFrame({
        'Hexspeak': (
            0xCAFED00D,
            0xCAFEBABE,
            0x1337BABE,
        ),
        'OdeToTheWestWind': (
            'Angels of rain and lightning: there are spread',
            'On the blue surface of thine aery surge,',
            'Like the bright hair uplifted from the head',
        ),
        'PercyByssheShelley': to_datetime((
            '1792-08-04',
            '1822-07-08',
            '1851-02-01',
        )),
    })

    # Pandas data frame containing arbitrary sample data violating the
    # "PanderaModel" defined below.
    pandas_dataframe_bad = DataFrame()

    # ....................{ SUBCLASSES                     }....................
    class PanderaModel(PandasDataFrameModel):
        '''
        Pandera model validating the ``pandas_dataframe_good`` defined above.
        '''

        Hexspeak: Series[Int64]
        OdeToTheWestWind: Series[String]
        PercyByssheShelley: Series[Timestamp]

    #FIXME: Preserved for debuggability. Uncomment as needed, please.
    # hint = PanderaDataFrame[PanderaModel]
    # hint_origin = hint.__origin__
    # # print(f'\nrepr(hint): {repr(hint)}')
    # # print(f'\ndir(hint): {dir(hint)}')
    # # print(f'\ntype(hint): {type(hint)}')
    # # from pandera.typing.pandas import DataFrame as NestedPanderaDataFrame
    # # print(f'\nis hint_origin a pandera dataframe? {hint_origin is NestedPanderaDataFrame}')
    # # print(f'\nrepr(hint_origin): {repr(hint_origin)}')
    # # print(f'\ndir(hint_origin): {dir(hint_origin)}')
    # print(f'\nmro(hint_origin): {hint_origin.__mro__}')
    # # print(f'\ntype(hint_origin): {type(hint_origin)}')
    # print(f'\nhint_origin.__orig_bases__: {hint_origin.__orig_bases__}')
    # print(f'is pandas dataframe a pandera dataframe? {isinstance(pandas_dataframe_good, hint_origin)}')
    # print(f'is pandas dataframe an orig dataframe? {isinstance(pandas_dataframe_good, hint_origin.__orig_bases__)}')
    # # print(f'is pandas dataframe a pandera dataframe? {isinstance(pandas_dataframe_good, NestedPanderaDataFrame)}')
    # # print(f'is pandas dataframe an orig dataframe? {isinstance(pandas_dataframe_good, NestedPanderaDataFrame.__orig_bases__)}')
    # from beartype import BeartypeConf
    # @beartype(conf=BeartypeConf(is_debug=True))

    # ....................{ FUNCTIONS                      }....................
    @beartype
    @pandas_check_types
    def to_the_zeniths_height(
        dataframe: PanderaDataFrame[PanderaModel],
        of_some_fierce_maenad: str,
    ) -> str:
        '''
        Arbitrary callable decorated first by :mod:`beartype` and then by
        :mod:`pandera`, accepting a parameter annotated by a Pandera type hint
        *and* a parameter annotated by a non-Pandera PEP-compliant type hint.

        This callable exercises that:

        * The :func:`pandera.check_types` decorator type-checks the parameter
          annotated by the Pandera type hint.
        * The :func:`beartype.beartype` decorator type-checks the parameter
          annotated by the non-Pandera type hint.
        '''

        # Return this string parameter appended by an arbitrary constant.
        return f'{of_some_fierce_maenad}, even from the dim verge'
    # print(f'PanderaDataFrame[pandera_schema]: {repr(PanderaDataFrame[pandera_schema])}')


    @pandas_check_types
    @beartype
    def all_thy_congregated_might(
        dataframe: PanderaDataFrame[PanderaModel],
        of_some_fierce_maenad: str,
    ) -> str:
        '''
        Arbitrary callable decorated first by :mod:`pandera` and then by
        :mod:`beartype`, accepting a parameter annotated by a Pandera type hint
        *and* a parameter annotated by a non-Pandera PEP-compliant type hint.

        This callable exercises that order of decoration is insignificant.
        '''

        # Return this string parameter appended by an arbitrary constant.
        return f'{of_some_fierce_maenad}, even from the dim verge'


    @beartype
    def from_whose_solid_atmosphere(
        dataframe: PanderaDataFrame[PanderaModel],
        of_some_fierce_maenad: str,
    ) -> str:
        '''
        Arbitrary callable decorated by :mod:`beartype` but *not* by
        :mod:`pandera`, accepting a parameter annotated by a Pandera type hint
        *and* a parameter annotated by a non-Pandera PEP-compliant type hint.

        This callable exercises that the :func:`beartype.beartype` decorator at
        least shallowly type-checks all passed parameters regardless of whether
        this callable is also decorated by the :func:`pandera.check_types`
        decorator.
        '''

        # Return this string parameter appended by an arbitrary constant.
        return f'{of_some_fierce_maenad}, even from the dim verge'


    # Tuple of all functions defined above decorated by @pandera.check_types.
    TEST_FUNCS_PANDERA = (
        to_the_zeniths_height,
        all_thy_congregated_might,
    )

    # Tuple of all functions defined above.
    TEST_FUNCS = TEST_FUNCS_PANDERA + (
        from_whose_solid_atmosphere,
    )

    # ....................{ ASSERTS                        }....................
    # For each function defined above to be tested...
    for test_func in TEST_FUNCS:
        # ....................{ PASS                       }....................
        # Assert that calling this function with valid parameters returns the
        # expected value.
        assert test_func(
            dataframe=pandas_dataframe_good,
            of_some_fierce_maenad='Of some fierce Maenad',
        ) == 'Of some fierce Maenad, even from the dim verge'

        # ....................{ FAIL                       }....................
        # Assert that calling this function with an invalid parameter managed by
        # @beartype raises the expected @beartype exception.
        with raises(BeartypeCallHintParamViolation):
            test_func(
                dataframe=pandas_dataframe_good,
                of_some_fierce_maenad=b'The locks of the approaching storm.',
            )

        # Assert that calling this function with an invalid parameter shallowly
        # type-checked by @beartype raises an... exception.
        #
        # Currently, Pandera currently fails to perform similar shallow
        # type-checking. Instead, Pandera silently ignores parameters that are
        # *NOT* of the expected type. Since this behaviour is both bizarre and
        # erroneous, future releases of Pandera are likely to at least shallowly
        # type-check parameters annotated by Pandera type hints. For forward
        # compatibility, we avoid asserting the exact type of this exception.
        # with raises(Exception):
        with raises(BeartypeCallHintParamViolation):
            test_func(
                dataframe='Black rain, and fire, and hail will burst: oh hear!',
                of_some_fierce_maenad=b'The locks of the approaching storm.',
            )

    # ....................{ ASSERTS ~ pandera              }....................
    # For each function defined above decorated by @pandera.check_types...
    for test_func_pandera in TEST_FUNCS_PANDERA:
        # ....................{ FAIL                       }....................
        # Assert that calling this function with an invalid parameter managed by
        # Pandera raises the expected Pandera exception.
        with raises(SchemaError):
            test_func_pandera(
                dataframe=pandas_dataframe_bad,
                of_some_fierce_maenad='Of some fierce Maenad',
            )

    # ....................{ MORE                           }....................
    # Functions defined above suffice for the trivial case of type-checking a
    # data frame passed as a parameter. Let's generalize that with an additional
    # function also type-checking a series returned as a value.

    @beartype
    @pandas_check_types
    def convert_dataframe_column_to_series(
        dataframe: PanderaDataFrame[PanderaModel],
        column_name_or_index: str | int,
    ) -> Series[Int64 | String | Timestamp]:
        '''
        Convert the column of the passed pandas data frame (identified by the
        passed column name or index) into a pandas series.
        '''

        # Return either...
        return (
            # If the caller passed a column name, the non-series column with
            # this name.
            #
            # Note that columns are *NOT* series; ergo, this return value
            # intentionally violates this return type hint. Doing so enables us
            # to assert that @beartype correctly type-checks return violations
            # involving pandera type hints.
            dataframe.loc
            # dataframe.loc[:,column_name_or_index]
            if isinstance(column_name_or_index, str) else
            # Else, the caller passed a column index. In this case, the series
            # converted from the column with this name.
            #
            # Note that this return value intentionally satisfies this hint.
            dataframe.iloc[:,column_name_or_index]
        )

    # Assert that calling this function with valid parameters but returning an
    # invalid value raises the expected exception.
    with raises(BeartypeCallHintReturnViolation):
        convert_dataframe_column_to_series(
            dataframe=pandas_dataframe_good, column_name_or_index='Hexspeak')

    # Assert that calling this function with valid parameters and returning a
    # valid value returns the expected object.
    pandas_series = convert_dataframe_column_to_series(
        dataframe=pandas_dataframe_good, column_name_or_index=0)
    assert len(pandas_series) == 3

# ....................{ TESTS ~ polars                     }....................
@skip_unless_package('pandera')
@skip_unless_package('polars')
# Prevent pytest from capturing and displaying non-fatal deprecation warnings
# emitted by Polars from pandera code. See also this relevant upstream issue:
#     https://github.com/unionai-oss/pandera/issues/2104
@ignore_warnings(DeprecationWarning)
def test_pandera_polars() -> None:
    '''
    Integration test validating that the :func:`beartype.beartype` decorator
    integrates cleanly with the third-party :class:`pandera.check_types`
    decorator with respect to Polars dataframes.

    Specifically, this test validates that the :func:`beartype.beartype`
    decorator conditionally ignores *all* Pandera type hints.

    See Also
    --------
    https://github.com/beartype/beartype/issues/557
        Feature request strongly inspiring this test.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from pandera.errors import SchemaError
    from pandera.polars import (
        DataFrameModel as PolarsDataFrameModel,
        Field as PolarsField,
        check_types as polars_check_types,
    )
    from pandera.typing.polars import DataFrame as PolarsDataFrameType
    from polars import DataFrame as PolarsDataFrame
    from pytest import raises

    # ....................{ SUBCLASSES                     }....................
    class PolarsDataFrameSchema(PolarsDataFrameModel):
        '''
        Arbitrary Pandera-specific schema validating Polars-specific data frames
        containing the columns and fields specified by this schema.
        '''

        smoothest_silence: str
        solemn_tubes: str
        serious_zephyrs: int = PolarsField(
            in_range={'min_value': 5, 'max_value': 20})

    # ....................{ FUNCTIONS                      }....................
    @polars_check_types
    @beartype
    def check_polars_dataframe(polars_dataframe: PolarsDataFrame) -> (
        PolarsDataFrameType[PolarsDataFrameSchema]):
        '''
        Arbitrary function annotated as accepting a Polars data frame and
        returning that same frame validated by the Pandera data frame schema
        defined above, decorated by both the :func:`beartype.beartype` and
        :func:`pandera.polars.check_types` decorators.
        '''

        return polars_dataframe

    # ....................{ LOCALS                         }....................
    # Polars data frame satisfying the Pandera data frame schema defined above.
    polars_dataframe_good = PolarsDataFrame({
        'smoothest_silence': [
            'In',   'In',   'In',     'smoothest', 'silence',],
        'solemn_tubes': [
            'save', 'what', 'solemn', 'tubes',     'Blown',],
        'serious_zephyrs': [
            10,     20,     15,       14,          12,]
    })

    # Polars data frame violating the Pandera data frame schema defined above.
    polars_dataframe_bad = PolarsDataFrame({
        'smoothest_silence': [
            'In',   'In',   'In',     'smoothest', 'silence',],
        'solemn_tubes': [
            'save', 'what', 'solemn', 'tubes',     'Blown',],
        'serious_zephyrs': [
            10,     20,     15,       14,          0,]
                                                   # ^-- *VIOLATION*!
    })

    # ....................{ PASS                           }....................
    # Implicitly assert that the above function accepts a Polars data frame
    # satisfying the Pandera data frame schema defined above.
    check_polars_dataframe(polars_dataframe_good)

    # ....................{ FAIL                           }....................
    # Assert that the above function rejects a Polars data frame violating the
    # Pandera data frame schema defined above by raising the expected exception.
    with raises(SchemaError):
        check_polars_dataframe(polars_dataframe_bad)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **import hook state** (i.e., data class singletons safely centralizing
*all* global state maintained by beartype import hooks, enabling each external
unit test in our test suite to trivially reset that state after completion of
that test).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.claw._ast._scope.clawastscopebefore import (
    BeartypeNodeScopeBeforelist,
    make_node_scope_beforelist_global,
)
from beartype.claw._importlib.clawimpcache import ModuleNameToBeartypeConf
from beartype.claw._package.clawpkgtrie import (
    PackagesTrieBlacklist,
    PackagesTrieBlacklisted,
    PackagesTrieWhitelist,
    PackageBasenameToTrieBlacklist,
)
from beartype._data.conf.dataconfblack import (
    BLACKLIST_PACKAGE_NAMES)
from beartype.typing import (
    TYPE_CHECKING,
    Optional,
)
from beartype._data.typing.datatyping import ImportPathHook
from threading import RLock

# ....................{ CLASSES                            }....................
class BeartypeClawState(object):
    '''
    **Beartype import hook state** (i.e., non-thread-safe singleton safely
    centralizing global state maintained by beartype import hooks, enabling each
    external unit test in our test suite to trivially reset that state after
    completion of that test).

    Attributes
    ----------
    beartype_pathhook : Optional[ImportPathHook]
        Either:

        * If the
          :func:`beartype.claw._importlib.clawimpmain.add_beartype_pathhook`
          function has been previously called at least once under the active
          Python interpreter and the
          :func:`beartype.claw._importlib.clawimpmain.remove_beartype_pathhook`
          function has not been called more recently, the **beartype import path
          hook singleton** (i.e., factory closure creating and returning a new
          :class:`importlib.machinery.FileFinder` instance itself creating and
          leveraging a new :class:`.BeartypeSourceFileLoader` instance).
        * Else, :data:`None`.

        Initialized to :data:`None`.
    module_name_to_beartype_conf : ModuleNameToBeartypeConf
        **Hooked module beartype configuration cache** (i.e., non-thread-safe
        dictionary mapping from the fully-qualified name of each previously
        imported submodule of each package previously registered in our global
        package trie to the beartype configuration configuring type-checking by
        the :func:`beartype.beartype` decorator of that submodule).
    node_scope_beforelist_global : BeartypeNodeScopeBeforelist
        **Abstract syntax tree (AST) global scope beforelist** (i.e., low-level
        dataclass aggregating all metadata required to manage the beforelist
        automating decorator positioning across all global scopes of all modules
        recursively visited by :mod:`beartype.claw` AST transformers).
    packages_trie_blacklist : PackagesTrieWhitelist
        **Package trie blacklist** (i.e., non-thread-safe recursively nested
        dictionary implementing a prefix tree such that each key-value pair maps
        from the unqualified basename of each subpackage to *not* be implicitly
        type-checked on the first importation of that subpackage to another
        instance of the :class:`.PackagesTrieWhitelist` class similarly
        describing the sub-subpackages of that subpackage).
    packages_trie_whitelist : PackagesTrieWhitelist
        **Package trie whitelist** (i.e., non-thread-safe recursively nested
        dictionary implementing a prefix tree such that each key-value pair maps
        from the unqualified basename of each subpackage to be implicitly
        type-checked on the first importation of that subpackage to another
        instance of the :class:`.PackagesTrieWhitelist` class similarly
        describing the sub-subpackages of that subpackage).
    '''

    # ..................{ CLASS VARIABLES                    }..................
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: When adding a new slot, ensure that the copy_deep() method
    # defined below deeply copies that slot as well.
    # CAUTION: Subclasses declaring uniquely subclass-specific instance
    # variables *MUST* additionally slot those variables. Subclasses violating
    # this constraint will be usable but unslotted, which defeats our purposes.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Slot all instance variables defined on this object to reduce the costs of
    # both reading and writing these variables by approximately ~10%.
    __slots__ = (
        'beartype_pathhook',
        'module_name_to_beartype_conf',
        'node_scope_beforelist_global',
        'packages_trie_blacklist',
        'packages_trie_whitelist',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        beartype_pathhook: Optional[ImportPathHook]
        module_name_to_beartype_conf: ModuleNameToBeartypeConf
        node_scope_beforelist_global: BeartypeNodeScopeBeforelist
        packages_trie_blacklist: PackagesTrieBlacklist
        packages_trie_whitelist: PackagesTrieWhitelist

    # ....................{ INITIALIZERS                   }....................
    def __init__(self) -> None:

        # Nullify the proper subset of instance variables requiring
        # nullification *BEFORE* reinitializing this singleton.
        self.beartype_pathhook: Optional[ImportPathHook] = None

        # Reinitialize this singleton safely.
        self._reinit_safe()


    def _reinit_safe(self) -> None:
        '''
        Reinitialize *all* beartype import hook state encapsulated by this data
        class back to their initial defaults, trivially clearing *all* metadata
        pertaining to previously hooked packages and configurations installed by
        previously called beartype import hooks.

        This method performs the subset of reinitialization that is safe to be
        called from the :meth:`__init__` method.
        '''

        # One one-liner to reinitialize them all.
        self.module_name_to_beartype_conf = ModuleNameToBeartypeConf()
        self.node_scope_beforelist_global = make_node_scope_beforelist_global()
        self.packages_trie_whitelist = PackagesTrieWhitelist()
        self.packages_trie_blacklist = PackagesTrieBlacklist(
            subpackage_basename_to_trie=_PACKAGE_NAME_TO_TRIE_BLACKLISTED)
        # print(f'node_scope_beforelist_global: {self.node_scope_beforelist_global}')

        #FIXME: Preserved because the above will inevitably break. *sigh*
        # self.packages_trie_blacklist = PackagesTrieBlacklist()


    def reinit(self) -> None:
        '''
        Reinitialize *all* beartype import hook state encapsulated by this data
        class back to their initial defaults, trivially clearing *all* metadata
        pertaining to previously hooked packages and configurations installed by
        previously called beartype import hooks.
        '''
        # print('Renitializing "beartype.claw" state...')

        # Avoid circular import dependencies.
        from beartype.claw._importlib.clawimpmain import (
            remove_beartype_pathhook)

        # Perform the subset of reinitialization that is safe to be called from
        # the __init__() method.
        self._reinit_safe()

        # Perform the remainder of reinitialization that is unsafe to be called
        # from the __init__() method.
        #
        # Remove our beartype import path hook if this path hook has already
        # been added (e.g., by a prior call to an import hook) *OR* silently
        # reduce to a noop otherwise.
        remove_beartype_pathhook()

    # ..................{ COPIERS                            }..................
    #FIXME: Unit test us up, please.
    #FIXME: Comment out all of this for the moment, please. That includes the
    #copy_deep() methods implemented below as well. They're not implemented
    #correctly at the moment, sadly. They need to call themselves recursively.
    #They don't. Thus, we all sigh. *sigh*
    # def copy_deep(self) -> 'BeartypeClawState':
    #     '''
    #     Deep copy of this beartype import hook state.
    #     '''
    #
    #     # New initially empty beartype import hook state.
    #     claw_state_copy = BeartypeClawState()
    #
    #     #FIXME: Deeply copy all of the remaining stuff, including:
    #     #* "module_name_to_beartype_conf".
    #     #* "node_scope_beforelist_global".
    #
    #     # Deeply copy *ALL* container-specific instance variables necessitating
    #     # a deep copy from the current beartype import hook state into this
    #     # copy.
    #     claw_state_copy.packages_trie_blacklist = (
    #         self.packages_trie_blacklist.copy_deep())
    #     claw_state_copy.packages_trie_whitelist = (
    #         self.packages_trie_whitelist.copy_deep())
    #
    #     # Shallowly copy *ALL* remaining instance variables.
    #     claw_state_copy.beartype_pathhook = self.beartype_pathhook
    #
    #     # Return this deep copy.
    #     return claw_state_copy

    # ..................{ DUNDERS                            }..................
    def __repr__(self) -> str:

        return '\n'.join((
            f'{self.__class__.__name__}(\n',
            f'    beartype_pathhook={repr(self.beartype_pathhook)},\n',
            f'    module_name_to_beartype_conf={repr(self.module_name_to_beartype_conf)},\n',
            f'    node_scope_beforelist_global={repr(self.node_scope_beforelist_global)},\n',
            f'    packages_trie_blacklist={repr(self.packages_trie_blacklist)},\n',
            f'    packages_trie_whitelist={repr(self.packages_trie_whitelist)},\n',
            f')',
        ))

# ....................{ PRIVATE ~ constants                }....................
# Fully initialized by the _init() function called below.
_PACKAGE_NAME_TO_TRIE_BLACKLISTED: PackageBasenameToTrieBlacklist = {}
'''
Dictionary mapping from the unqualified basename of each top-level package to be
blacklisted (i.e., prevented from being runtime type-checked on the first
importation of that package) to the :data:`.PackagesTrieBlacklisted` object
similarly blacklisting the subpackages of that package if any.
'''

# ....................{ PRIVATE ~ initializers             }....................
def _init() -> None:
    '''
    Initialize this submodule.
    '''

    # Set of the names of all top-level packages to be blacklisted with respect
    # to "beartype.claw" import hooks -- defined as the union of:
    # * The frozen set of the names all third-party packages to be unilaterally
    #   blacklisted across this entire codebase.
    # * The root "beartype" package. Doing so effectively silently ignores
    #   dangerous attempts to recursively type-check the "beartype" package by
    #   the @beartype.beartype decorator. See the
    #   beartype.claw._importlib._clawimpload.BeartypeSourceFileLoader.get_code()
    #   method docstring for further commentary.
    #
    # Note that "beartype" is intentionally *OMITTED* from the global
    # "BLACKLIST_PACKAGE_NAMES" frozen set iterated over below. Why?
    # Because "beartype" should *ONLY* be blacklisted with respect to
    # "beartype.claw" import hooks. "beartype" should *NOT* be unilaterally
    # blacklisted across the entirety of this codebase, as doing so would
    # erroneously destroy our ability to (in no particular order):
    # * Define deeply type-checkable PEP 484- and 585-compliant generics.
    package_names_blacklist = BLACKLIST_PACKAGE_NAMES | {
        'beartype',
    }

    # For the name of each top-level package to be blacklisted with respect to
    # "beartype.claw" import hooks...
    for package_name_blacklist in package_names_blacklist:
        # Validate this package to actually be top-level.
        assert '.' not in package_name_blacklist, (
            f'{repr(package_name_blacklist)} not top-level package name '
            f'(i.e., contains one or more "." delimiters).'
        )

        # Map this package name to the "PackagesTrieBlacklist" singleton
        # implying this package to be blacklisted.
        _PACKAGE_NAME_TO_TRIE_BLACKLISTED[package_name_blacklist] = (
            PackagesTrieBlacklisted)


# Initialize this submodule.
_init()

# ....................{ GLOBALS                            }....................
# These globals require this submodule to be fully initialized and are thus
# intentionally defined *AFTER* all other code above. We sigh, fam. *sigh*

claw_lock = RLock()
'''
Reentrant reusable thread-safe context manager gating access to the otherwise
non-thread-safe :data:`.claw_state` global.
'''


claw_state = BeartypeClawState()
'''
**Beartype import hook state** (i.e., non-thread-safe singleton safely
centralizing *all* global state maintained by beartype import hooks, enabling
each external unit test in our test suite to trivially reset that state after
completion of that test).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`484`- or :pep:`585`-compliant **generic type-checking code
factories** (i.e., low-level callables dynamically generating pure-Python code
snippets type-checking arbitrary objects against generic type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._check.metadata.hint.hintsmeta import HintsMeta
from beartype._check.pep.checkpep484585generic import (
    get_hint_pep484585_generic_unsubbed_bases_unerased)
from beartype._data.code.pep.datacodepep484585 import (
    CODE_PEP484585_GENERIC_CHILD_format,
    CODE_PEP484585_GENERIC_PREFIX,
    CODE_PEP484585_GENERIC_SUFFIX,
)
from beartype._data.code.datacodelen import LINE_RSTRIP_INDEX_AND
from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
    get_hint_pep484585_generic_type_isinstanceable)

# ....................{ FACTORIES                          }....................
def make_hint_pep484585_generic_unsubbed_check_expr(
    hints_meta: HintsMeta) -> None:
    '''
    Either a Python code snippet type-checking the current pith against the
    passed :pep:`484`- or :pep:`585`-compliant **unsubscripted generic,**
    defined as either:

    * :pep:`484`-compliant **unsubscripted generic** (i.e., user-defined class
      subclassing a combination of one or more of the :class:`typing.Generic`
      superclass and other :mod:`typing` non-class pseudo-superclasses) *or*...
    * :pep:`544`-compliant **unsubscripted protocol** (i.e., class subclassing a
      combination of one or more of the :class:`typing.Protocol` superclass and
      other :mod:`typing` non-class pseudo-superclasses) *or*...
    * :pep:`585`-compliant unsubscripted generic (i.e., user-defined class
      subclassing at least one non-class :pep:`585`-compliant
      pseudo-superclasses).

    This factory is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), as the ``hints_meta`` parameter is
    **context-sensitive** (i.e., contextually depends on context unique to the
    code being generated for the currently decorated callable).

    Parameters
    ----------
    hints_meta : HintsMeta
        Stack of metadata describing all visitable hints previously discovered
        by this breadth-first search (BFS).
    '''
    assert isinstance(hints_meta, HintsMeta), (
        f'{repr(hints_meta)} not "HintsMeta" object.')
    # print(f'Visiting generic type {repr(hint_curr)}...')

    # ....................{ LOCALS                         }....................
    # Metadata encapsulating the sanification of this unsubscripted generic,
    # localized for both usability and efficiency.
    hint_sane = hints_meta.hint_curr_meta.hint_sane

    # Unsubscripted generic encapsulated by this metadata.
    hint = hint_sane.hint

    # Isinstanceable type against which to type-check instances of this generic,
    # defaulting to this generic. Although most generics are isinstanceable,
    # some are not. This type enables this code generator to transparently
    # support the subset of generics that are *NOT* isinstanceable.
    hint_isinstanceable = get_hint_pep484585_generic_type_isinstanceable(
        hint=hint, exception_prefix=hints_meta.exception_prefix)

    # ....................{ FORMAT                         }....................
    # Initialize the code type-checking this pith against this generic to the
    # substring prefixing all such code.
    hints_meta.func_curr_code = CODE_PEP484585_GENERIC_PREFIX

    # For metadata encapsulating the sanification of each unignorable unerased
    # transitive pseudo-superclass originally declared as a superclass of this
    # unsubscripted generic *AND* the sign identifying this pseudo-superclass...
    for hint_child_sane, hint_child_sign in (
        get_hint_pep484585_generic_unsubbed_bases_unerased(
            hint_sane,
            hints_meta.cls_stack,
            hints_meta.conf,
            hints_meta.exception_prefix,
        )
    ):
        # print(f'Visiting generic type hint {hint_curr_sane} unerased base {hint_child_sane}...')

        # Append code type-checking this pith against this pseudo-superclass.
        hints_meta.func_curr_code += CODE_PEP484585_GENERIC_CHILD_format(
            hint_child_placeholder=hints_meta.enqueue_hint_child_sane(
                hint_sane=hint_child_sane,
                hint_sign=hint_child_sign,
                # Python expression efficiently reusing the value of this pith
                # previously assigned to a local variable by the prior
                # expression.
                pith_expr=hints_meta.pith_curr_var_name,
            ),
        )

    # Munge this code to...
    hints_meta.func_curr_code = (
        # Strip the erroneous " and" suffix appended by the last child hint from
        # this code.
        f'{hints_meta.func_curr_code[:LINE_RSTRIP_INDEX_AND]}'
        # Suffix this code by the substring suffixing all such code.
        f'{CODE_PEP484585_GENERIC_SUFFIX}'
    # Format...
    ).format(
        # Indentation deferred above for efficiency.
        indent_curr=hints_meta.indent_curr,
        pith_curr_assign_expr=hints_meta.pith_curr_assign_expr,
        # Python expression evaluating to this unsubscripted isinstanceable
        # generic type.
        hint_curr_expr=hints_meta.add_func_scope_type_or_types(
            hint_isinstanceable),
    )
    # print(f'{hint_curr_exception_prefix} PEP generic {repr(hint)} handled.')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **object utilities** (i.e., low-level callables handling arbitrary
objects in a general-purpose manner).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilObjectNameException
from beartype.typing import (
    Any,
    Optional,
)
from beartype._data.cls.datacls import TYPES_CONTEXTMANAGER_FAKE
from contextlib import AbstractContextManager

# ....................{ TESTERS                            }....................
def is_object_context_manager(obj: object) -> bool:
    '''
    :data:`True` only if the passed object is a **context manager** (i.e.,
    object defining both the ``__exit__`` and ``__enter__`` dunder methods
    required to satisfy the context manager protocol).

    Parameters
    ----------
    obj : object
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a context manager.
    '''

    # Return true only if...
    return (
        # This object satisfies the context manager protocol (i.e., defines both
        # the __enter__() and __exit__() dunder methods) *AND*...
        isinstance(obj, AbstractContextManager) and
        # This object is *NOT* a "fake" context manager (i.e., defines erroneous
        # __enter__() and __exit__() dunder methods trivially reducing to noops
        # and also emitting non-fatal deprecation warnings).
        not isinstance(obj, TYPES_CONTEXTMANAGER_FAKE)
    )


# Note that this tester function *CANNOT* be memoized by the @callable_cached
# decorator, which requires all passed parameters to already be hashable.
def is_object_hashable(obj: object) -> bool:
    '''
    :data:`True` only if the passed object is **hashable** (i.e., passable to
    the builtin :func:`hash` function *without* raising an exception and thus
    usable in hash-based containers like dictionaries and sets).

    Parameters
    ----------
    obj : object
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is hashable.
    '''

    # Attempt to hash this object. If doing so raises *any* exception
    # whatsoever, this object is by definition unhashable.
    #
    # Note that there also exists a "collections.abc.Hashable" superclass.
    # Sadly, this superclass is mostly useless for all practical purposes. Why?
    # Because user-defined classes are free to subclass that superclass
    # despite overriding the __hash__() dunder method implicitly called by the
    # builtin hash() function to raise exceptions: e.g.,
    #
    #     from collections.abc import Hashable
    #     class HashUmUp(Hashable):
    #         def __hash__(self):
    #             raise ValueError('uhoh')
    #
    # Note also that we catch all possible exceptions rather than merely the
    # standard "TypeError" exception raised by unhashable builtin types (e.g.,
    # dictionaries, lists, sets). Why? For the same exact reason as above.
    try:
        hash(obj)
    # If this object is unhashable, return false.
    except Exception:
        return False

    # Else, this object is hashable. Return true.
    return True

# ....................{ GETTERS ~ name                     }....................
def get_object_name(obj: Any) -> str:
    '''
    **Fully-qualified name** (i.e., ``.``-delimited string unambiguously
    identifying) of the passed object if this object defines either the
    ``__qualname__`` or ``__name__`` dunder attributes *or* raise an exception
    otherwise (i.e., if this object defines *no* such attributes).

    Specifically, this name comprises (in order):

    #. If this object is transitively declared by a module, the absolute name
       of that module.
    #. If this object is transitively declared by another object (e.g., class,
       callable) and thus nested in that object, the unqualified basenames of
       all parent objects transitively declaring this object in that module.
    #. Unqualified basename of this object.

    Parameters
    ----------
    obj : object
        Object to be inspected.

    Returns
    -------
    str
        Fully-qualified name of this object.

    Raises
    ------
    _BeartypeUtilObjectNameException
        If this object defines neither ``__qualname__`` *nor* ``__name__``
        dunder attributes.
    '''

    # Avoid circular import dependencies.
    from beartype._cave._cavefast import CallableOrClassTypes
    from beartype._util.module.utilmodget import (
        get_object_module_name_or_none,
        get_object_type_module_name_or_none,
    )

    # Lexically scoped name of this object excluding this module name if this
    # object is named *OR* raise an exception otherwise.
    object_scopes_name = get_object_basename_scoped(obj)

    # Fully-qualified name of the module declaring this object if this object
    # is declared by a module *OR* "None" otherwise, specifically defined as:
    # * If this object is either a callable or class, the fully-qualified name
    #   of the module declaring this object.
    # * Else, the fully-qualified name of the module declaring the class of this
    #   object.
    object_module_name = (
        get_object_module_name_or_none(obj)
        if isinstance(obj, CallableOrClassTypes) else
        get_object_type_module_name_or_none(obj)
    )

    # Return either...
    return (
        # If this module name exists, "."-delimited concatenation of this module
        # and object name;
        f'{object_module_name}.{object_scopes_name}'
        if object_module_name is not None else
        # Else, this object name as is.
        object_scopes_name
    )

# ....................{ GETTERS ~ basename                 }....................
def get_object_basename_scoped(obj: Any) -> str:
    '''
    **Lexically scoped name** (i.e., ``.``-delimited string unambiguously
    identifying all lexical scopes encapsulating) the passed object if this
    object defines either the ``__qualname__`` or ``__name__`` dunder attributes
    *or* raise an exception otherwise (i.e., if this object defines *no* such
    attributes).

    Parameters
    ----------
    obj : object
        Object to be inspected.

    Returns
    -------
    str
        Lexically scoped name of this object.

    Raises
    ------
    _BeartypeUtilObjectNameException
        If this object defines neither ``__qualname__`` *nor* ``__name__``
        dunder attributes.

    See Also
    --------
    :func:`.get_object_basename_scoped_or_none`
        Further details.
    '''

    # Fully-qualified name of this object excluding its module name.
    object_scoped_name = get_object_basename_scoped_or_none(obj)

    # If this object is unnamed, raise a human-readable exception. The default
    # "AttributeError" exception raised by attempting to directly access either
    # the "obj.__name__" or "obj.__qualname__" attributes is sufficiently
    # non-explanatory to warrant replacement by our explanatory exception.
    if object_scoped_name is None:
        raise _BeartypeUtilObjectNameException(
            f'{repr(obj)} unnamed '
            f'(i.e., declares neither "__name__" nor "__qualname__" '
            f'dunder attributes).'
        )
    # Else, this object is named.

    # Remove all "<locals>" placeholder substrings as discussed above.
    return object_scoped_name.replace('<locals>.', '')


#FIXME: Unit test us up, please.
def get_object_basename_scoped_or_none(obj: Any) -> Optional[str]:
    '''
    **Lexically scoped name** (i.e., ``.``-delimited string unambiguously
    identifying all lexical scopes encapsulating) the passed object if this
    object defines either the ``__qualname__`` or ``__name__`` dunder attributes
    *or* :data:`None` otherwise (i.e., if this object defines *no* such
    attributes).

    Specifically, this name comprises (in order):

    #. If this object is transitively declared by another object (e.g., class,
       callable) and thus nested in that object, the unqualified basenames of
       all parent objects transitively declaring this object in that module.
       For usability, these basenames intentionally omit the meaningless
       placeholder ``"<locals>"`` substrings artificially injected by Python
       itself into the original ``__qualname__`` instance variable underlying
       this getter: e.g.,

       .. code-block:: python

          >>> from beartype._util.utilobject import get_object_basename_scoped
          >>> def muh_func():
          ...     def muh_closure(): pass
          ...     return muh_closure()
          >>> muh_func().__qualname__
          'muh_func.<locals>.muh_closure'  # <-- bad Python
          >>> get_object_basename_scoped(muh_func)
          'muh_func.muh_closure'  # <-- good @beartype

    #. Unqualified basename of this object.

    Caveats
    -------
    **The higher-level** :func:`get_object_name` **getter should typically be
    called instead of this lower-level getter.** This getter unsafely:

    * Requires the passed object to declare dunder attributes *not* generally
      declared by arbitrary instances of user-defined classes.
    * Omits the fully-qualified name of the module transitively declaring this
      object and thus fails to return fully-qualified names.

    **This high-level getter should always be called in lieu of directly
    accessing the low-level** ``__qualname__`` **dunder attribute on objects.**
    That attribute contains one meaningless ``"<locals>"`` placeholder
    substring conveying *no* meaningful semantics for each parent callable
    lexically nesting this object.

    Parameters
    ----------
    obj : object
        Object to be inspected.

    Returns
    -------
    Optional[str]
        Either:

        * If this object defines at least one of the ``__qualname__`` or
          ``__name__`` dunder attributes, the lexically scoped name of this
          object.
        * Else, :data:`None`.

    Raises
    ------
    _BeartypeUtilObjectNameException
        If this object defines neither ``__qualname__`` *nor* ``__name__``
        dunder attributes.
    '''

    # Fully-qualified name of this object excluding its module name as follows:
    # * If this object defines the "__qualname__" dunder attribute whose value
    #   is the "."-delimited concatenation of the unqualified basenames of all
    #   parent objects transitively declaring this object, that value with all
    #   meaningless "<locals>" placeholder substrings removed. If this object
    #   is a nested non-method callable (i.e., pure-Python function nested in
    #   one or more parent pure-Python callables), that value contains one such
    #   placeholder for each parent callable containing this callable. Since
    #   placeholders convey no meaningful semantics, placeholders are removed.
    # * Else if this object defines the "__name__" dunder attribute whose value
    #   is the unqualified basename of this object, that value.
    # * Else, "None".
    object_scoped_name = getattr(
        obj, '__qualname__', getattr(
            obj, '__name__', None))

    # Return either...
    return (
        # If this name exists, all "<locals>" placeholder substrings globally
        # removed from this name as discussed above;
        object_scoped_name.replace('<locals>.', '')
        if object_scoped_name else
        # Else, either "None" or the empty string.
        object_scoped_name
    )

# ....................{ GETTERS ~ filename                 }....................
def get_object_filename_or_none(obj: object) -> Optional[str]:
    '''
    Filename of the module or script physically declaring the passed object if
    this object is either a callable or class physically declared on-disk *or*
    :data:`None` otherwise (i.e., if this object is neither a callable nor
    class *or* is either a callable or class dynamically declared in-memory).

    Parameters
    ----------
    obj : object
        Object to be inspected.

    Returns
    -------
    Optional[str]
        Either:

        * If this object is either a callable or class physically declared
          on-disk, the filename of the module or script physically declaring
          this object.
        * Else, :data:`None`.
    '''

    # Avoid circular import dependencies.
    from beartype._util.cls.utilclsget import get_type_filename_or_none
    from beartype._util.func.utilfuncfile import get_func_filename_or_none
    from beartype._util.func.utilfunctest import is_func_codeobjable

    # Return either...
    return (
        # If this object is a pure-Python class, the absolute filename of the
        # source module file defining that class if that class was defined
        # on-disk *OR* "None" otherwise (i.e., if that class was defined
        # in-memory);
        get_type_filename_or_none(obj)
        if isinstance(obj, type) else
        # If this object is a pure-Python callable, the absolute filename of the
        # absolute filename of the source module file defining that callable if
        # that callable was defined on-disk *OR* "None" otherwise (i.e., if that
        # callable was defined in-memory);
        get_func_filename_or_none(obj)
        if is_func_codeobjable(obj) else
        # Else, "None".
        None
    )

# ....................{ GETTERS ~ type                     }....................
def get_object_type_unless_type(obj: object) -> type:
    '''
    Either the passed object if this object is a class *or* the class of this
    object otherwise (i.e., if this object is *not* a class).

    Note that this function *never* raises exceptions on arbitrary objects, as
    the :obj:`type` builtin wisely returns itself when passed itself: e.g.,

    .. code-block:: python

        >>> type(type(type)) is type
        True

    Parameters
    ----------
    obj : object
        Object to be inspected.

    Returns
    -------
    type
        Type of this object.
    '''

    return obj if isinstance(obj, type) else type(obj)

# ....................{ GETTERS ~ type : name              }....................
def get_object_type_basename(obj: object) -> str:
    '''
    **Unqualified name** (i.e., non-``.``-delimited basename) of either the
    passed object if this object is a class *or* the class of this object
    otherwise (i.e., if this object is *not* a class).

    Parameters
    ----------
    obj : object
        Object to be inspected.

    Returns
    -------
    str
        Unqualified name of this class.
    '''

    # Elegant simplicity diminishes aggressive tendencies.
    return get_object_type_unless_type(obj).__name__


def get_object_type_name(obj: object) -> str:
    '''
    **Fully-qualified name** (i.e., ``.``-delimited name prefixed by the
    declaring module) of either passed object if this object is a class *or*
    the class of this object otherwise (i.e., if this object is *not* a class).

    Parameters
    ----------
    obj : object
        Object to be inspected.

    Returns
    -------
    str
        Fully-qualified name of the type of this object.
    '''

    # Avoid circular import dependencies.
    from beartype._util.module.utilmodget import (
        get_object_type_module_name_or_none)

    # Type of this object.
    cls = get_object_type_unless_type(obj)

    # Unqualified name of this type.
    cls_basename = get_object_type_basename(cls)

    # Fully-qualified name of the module defining this class if this class is
    # defined by a module *OR* "None" otherwise.
    cls_module_name = get_object_type_module_name_or_none(cls)

    # Return either...
    return (
        # The "."-delimited concatenation of this class basename and module
        # name if this module name exists.
        f'{cls_module_name}.{cls_basename}'
        if cls_module_name is not None else
        # This class basename as is otherwise.
        cls_basename
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **forward reference testers** (i.e.,  low-level callables testing
various properties of forward reference proxy subclasses deferring the
resolution of a stringified type hint referencing an attribute that has yet to
be defined and annotating a class or callable decorated by the
:func:`beartype.beartype` decorator).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._check.forward.reference.fwdrefmeta import BeartypeForwardRefMeta

# ....................{ TESTERS                            }....................
def is_beartype_forwardref(obj: object) -> bool:
    '''
    :data:`True` only if the passed object is a **forward reference proxy**
    (i.e., :mod:`beartype`-specific class whose metaclass is
    class:`.BeartypeForwardRefMeta`).

    Caveats
    -------
    **This high-level tester should always be called in lieu of lower-level
    operations,** especially attempts to detect forward reference proxies by
    calling either the :func:`isinstance` or :func:`issubclass` builtins. Since
    forward reference proxies proxy calls to those builtins, forward reference
    proxies *cannot* be detected by calling either of those builtins.

    Parameters
    ----------
    obj : object
        Object to be tested.

    Returns
    -------
    bool
        :data:`True` only if this object is a forward reference subclass.
    '''

    # Return true only if the class of this object is the metaclass of all
    # forward reference subclasses, implying this object to be such a subclass.
    return obj.__class__ is BeartypeForwardRefMeta



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **abstract syntax tree (AST) getter utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.ast.utilastget` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ attr                       }....................
def test_get_node_attr_basename_first() -> None:
    '''
    Test the :func:`beartype._util.ast.utilastget.get_node_attr_basename_first`
    factory.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.ast.utilastget import get_node_attr_basename_first
    from beartype._util.ast.utilastmake import make_node_from_code_snippet

    # ....................{ LOCALS                         }....................
    # "ast.Name" node encapsulating the unqualified basename of an arbitrary
    # global attribute, parsed from a code snippet.
    node_name = make_node_from_code_snippet('''instead_of_sweets''')

    # "ast.Attribute" node encapsulating the fully-qualified basename of an
    # arbitrary global attribute, parsed from a code snippet.
    node_attr = make_node_from_code_snippet('''his_ample.palate_took''')

    # Arbitrary non-empty list to be passed to this getter below.
    attr_basenames = ['Savour of', 'poisonous brass', 'and metal sick:']

    # ....................{ ASSERT                         }....................
    # Assert this getter passed a "Name" node returns the expected string.
    assert get_node_attr_basename_first(node_name) == 'instead_of_sweets'

    # Assert this getter passed an "Attribute" node returns the expected string.
    assert get_node_attr_basename_first(node_attr) == 'his_ample'

    # Assert this getter passed both an "Attribute" node *AND* list returns the
    # expected string.
    assert get_node_attr_basename_first(
        node=node_attr, attr_basenames=attr_basenames) == 'his_ample'


def test_get_node_attr_basenames() -> None:
    '''
    Test the :func:`beartype._util.ast.utilastget.get_node_attr_basenames`
    factory.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.ast.utilastget import get_node_attr_basenames
    from beartype._util.ast.utilastmake import make_node_from_code_snippet

    # ....................{ LOCALS                         }....................
    # "ast.Name" node encapsulating the unqualified basename of an arbitrary
    # global attribute, parsed from a code snippet.
    node_name = make_node_from_code_snippet('''taste_the_spicy_wreaths''')

    # "ast.Attribute" node encapsulating the fully-qualified basename of an
    # arbitrary global attribute, parsed from a code snippet.
    node_attr = make_node_from_code_snippet('''of_incense.breathed.aloft''')

    # Arbitrary non-empty list to be passed to this getter below.
    attr_basenames = ['from', 'sacred hills']

    # ....................{ ASSERT                         }....................
    # Assert this getter passed a "Name" node returns the expected list.
    assert get_node_attr_basenames(node_name) == ['taste_the_spicy_wreaths']

    # Assert this getter passed an "Attribute" node returns the expected list.
    assert get_node_attr_basenames(node_attr) == [
        'of_incense', 'breathed', 'aloft']

    # ....................{ ASSERT ~ attr_basenames        }....................
    # List of the one or more unqualified basenames comprising the possibly
    # fully-qualified name of the passed "Attribute" node such that the returned
    # list is the passed "attr_basenames" list.
    attr_basenames_new = get_node_attr_basenames(
        node=node_attr, attr_basenames=attr_basenames)

    # Assert these two lists to indeed be the same.
    assert attr_basenames_new is attr_basenames

    # Assert this list to contain the expected basenames.
    assert attr_basenames == ['of_incense', 'breathed', 'aloft']



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-compliant PEP-noncompliant type hint test data.

:pep:`484`-compliant type hints *mostly* indistinguishable from
PEP-noncompliant type hints include:

* :func:`typing.NamedTuple`, a high-level factory function deferring to the
  lower-level :func:`collections.namedtuple` factory function creating and
  returning :class:`tuple` instances annotated by PEP-compliant type hints.
* :func:`typing.TypedDict`, a high-level factory function creating and
  returning :class:`dict` instances annotated by PEP-compliant type hints.
'''

# ....................{ FIXTURES                           }....................
def hints_nonpep484_meta() -> 'List[HintNonpepMetadata]':
    '''
    List of :pep:`484`-sorta-compliant **type hint metadata** (i.e.,
    :class:`beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintNonpepMetadata`
    instances describing test-specific :pep:`484`-sorta-compliant sample type
    hints with metadata generically leveraged by various PEP-agnostic unit
    tests).
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    import sys
    from beartype import BeartypeConf
    from beartype.typing import (
        NamedTuple,
    )
    from beartype._cave._cavefast import (
        EllipsisType,
        FunctionType,
        FunctionOrMethodCType,
        MethodBoundInstanceOrClassType,
        ModuleType,
        NoneType,
        NotImplementedType,
    )
    from beartype_test.a00_unit.data.data_type import Class
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintNonpepMetadata,
        HintPithSatisfiedMetadata,
        HintPithUnsatisfiedMetadata,
    )

    # ..................{ LOCALS                             }..................
    # List of all PEP-noncompliant type hint metadata to be returned.
    hints_nonpep_meta = []

    #FIXME: *WOOPS.* We should have read the standards a bit closer. Neither
    #"typing.NamedTuple" or "typing.TypedDict" are intended for direct use as
    #type hints. To quote official "typing" documentation:
    #    These are not used in annotations. They are building blocks for
    #    declaring types.
    #
    #Of course, all types *ARE* valid type hints. "typing.NamedTuple" and
    #"typing.TypedDict" subclasses are types and thus also valid type hints. So,
    #the superficial testing we perform below is certainly useful; we just don't
    #need to do anything further, really. Phew!

    # PEP-compliant user-defined "collections.namedtuple" instance typed with
    # PEP-compliant type hints.
    NamedTupleType = NamedTuple(
        'NamedTupleType', [('fumarole', str), ('enrolled', int)])

    # ..................{ LISTS                              }..................
    # Add PEP 484-specific (albeit technically PEP-noncompliant from the
    # beartype perspective) test type hints to this list.
    hints_nonpep_meta.extend((
        # ................{ NAMEDTUPLE                         }................
        # "NamedTuple" instances transparently reduce to standard tuples and
        # *MUST* thus be handled as non-"typing" type hints.
        HintNonpepMetadata(
            hint=NamedTupleType,
            piths_meta=(
                # Named tuple containing correctly typed items.
                HintPithSatisfiedMetadata(
                    NamedTupleType(fumarole='Leviathan', enrolled=37)),
                # String constant.
                HintPithUnsatisfiedMetadata('Of ͼarthen concordance that'),
                #FIXME: Uncomment after implementing "NamedTuple" support.
                # # Named tuple containing incorrectly typed items.
                # HintPithUnsatisfiedMetadata(
                #     pith=NamedTupleType(fumarole='Leviathan', enrolled=37),
                #     # Match that the exception message raised for this object...
                #     exception_str_match_regexes=(
                #         # Declares the name of this tuple's problematic item.
                #         r'\s[Ll]ist item 0\s',
                #     ),
                # ),
            ),
        ),

        # ................{ TYPEDDICT                          }................
        # "TypedDict" instances transparently reduce to dicts.
        #FIXME: Implement us up, but note when doing so that:
        #* We currently unconditionally reduce "TypeDict" to "Mapping".
        #* "TypedDict" was first introduced with Python 3.8.

        # ................{ TYPE ~ builtin                     }................
        # Integer.
        HintNonpepMetadata(
            hint=int,
            piths_meta=(
                # Integer constant.
                HintPithSatisfiedMetadata(42),  # <-- we went there, folks
                # String constant.
                HintPithUnsatisfiedMetadata(
                    pith='Introspectively ‘allein,’ dealigning consangui-',
                    # Match that the exception message raised for this pith
                    # contains...
                    exception_str_match_regexes=(
                        # The type *NOT* satisfied by this object.
                        r'\bint\b',
                    ),
                    # Match that the exception message raised for this pith does
                    # *NOT* contain...
                    exception_str_not_match_regexes=(
                        # A newline.
                        r'\n',
                        # A bullet delimiter.
                        r'\*',
                        # The double-quoted name of this builtin type.
                        r'"int"',
                    ),
                ),
            ),
        ),

        # Unicode string.
        HintNonpepMetadata(
            hint=str,
            piths_meta=(
                # String constant.
                HintPithSatisfiedMetadata('Glassily lassitudinal bȴood-'),
                # Byte-string constant.
                HintPithUnsatisfiedMetadata(
                    pith=b'Stains, disdain-fully ("...up-stairs!"),',
                    # Match that the exception message raised for this pith
                    # contains...
                    exception_str_match_regexes=(
                        # The type *NOT* satisfied by this object.
                        r'\bstr\b',
                        # The representation of this object preserved as is.
                        r'\sb\'Stains, disdain-fully \("...up-stairs!"\),\'\s',
                    ),
                    # Match that the exception message raised for this pith
                    # does *NOT* contain...
                    exception_str_not_match_regexes=(
                        # A newline.
                        r'\n',
                        # A bullet delimiter.
                        r'\*',
                        # The double-quoted name of this builtin type.
                        r'"str"',
                    ),
                ),
                # Integer constant.
                HintPithUnsatisfiedMetadata(
                    pith=666,  # <-- number of the beast, yo
                    # Match that the exception message raised for this pith
                    # contains...
                    exception_str_match_regexes=(
                        # The type *NOT* satisfied by this object.
                        r'\bstr\b',
                        # The representation of this object preserved as is.
                        r'\s666\s',
                    ),
                    # Match that the exception message raised for this pith
                    # does *NOT* contain...
                    exception_str_not_match_regexes=(
                        # A newline.
                        r'\n',
                        # A bullet delimiter.
                        r'\*',
                        # The double-quoted name of this builtin type.
                        r'"str"',
                    ),
                ),
            ),
        ),

        # ................{ TYPE ~ builtin : tower             }................
        # Types pertaining to the implicit numeric tower (i.e., optional PEP
        # 484-compliant (sub)standard in which type hints defined as broad
        # numeric types implicitly match all narrower numeric types as well by
        # enabling the "beartype.BeartypeConf.is_pep484_tower" parameter). When
        # enabled, @beartype implicitly expands:
        # * "float" to "float | int".
        # * "complex" to "complex | float | int".
        #
        # See also the "_data_pep484" submodule, which defines additional
        # PEP 484-compliant type hints nesting type hints pertaining to the
        # implicit numeric tower (e.g., "Union[float, str]").

        # Floating-point number with the implicit numeric tower disabled.
        HintNonpepMetadata(
            hint=float,
            conf=BeartypeConf(is_pep484_tower=False),
            piths_meta=(
                # Floating-point number constant.
                HintPithSatisfiedMetadata(0.110001),
                # Integer constant.
                HintPithUnsatisfiedMetadata(
                    pith=110001,
                    # Match that the exception message raised for this pith
                    # contains...
                    exception_str_match_regexes=(
                        # The type *NOT* satisfied by this object.
                        r'\bfloat\b',
                    ),
                    # Match that the exception message raised for this pith
                    # does *NOT* contain...
                    exception_str_not_match_regexes=(
                        # A newline.
                        r'\n',
                        # A bullet delimiter.
                        r'\*',
                        # The double-quoted name of this builtin type.
                        r'"float"',
                    ),
                ),
            ),
        ),

        # Floating-point number with the implicit numeric tower enabled.
        HintNonpepMetadata(
            hint=float,
            conf=BeartypeConf(is_pep484_tower=True),
            piths_meta=(
                # Floating-point number constant.
                HintPithSatisfiedMetadata(0.577215664901532860606512090082),
                # Integer constant.
                HintPithSatisfiedMetadata(5772),
                # Complex number constant.
                HintPithUnsatisfiedMetadata(
                    pith=(1566 + 4901j),
                    # Match that the exception message raised for this pith
                    # contains...
                    exception_str_match_regexes=(
                        # The type *NOT* satisfied by this object.
                        r'\bfloat\b',
                    ),
                    # Match that the exception message raised for this pith
                    # does *NOT* contain...
                    exception_str_not_match_regexes=(
                        # A newline.
                        r'\n',
                        # A bullet delimiter.
                        r'\*',
                        # The double-quoted name of this builtin type.
                        r'"float"',
                    ),
                ),
            ),
        ),

        # Complex number with the implicit numeric tower disabled.
        HintNonpepMetadata(
            hint=complex,
            conf=BeartypeConf(is_pep484_tower=False),
            piths_meta=(
                # Complex number constant.
                HintPithSatisfiedMetadata(1.787 + 2316.5j),
                # Floating-point number constant.
                HintPithUnsatisfiedMetadata(
                    pith=0.300330000000000330033,
                    # Match that the exception message raised for this pith
                    # contains...
                    exception_str_match_regexes=(
                        # The type *NOT* satisfied by this object.
                        r'\bcomplex\b',
                    ),
                    # Match that the exception message raised for this pith
                    # does *NOT* contain...
                    exception_str_not_match_regexes=(
                        # A newline.
                        r'\n',
                        # A bullet delimiter.
                        r'\*',
                        # The double-quoted name of this builtin type.
                        r'"complex"',
                    ),
                ),
            ),
        ),

        # Complex number with the implicit numeric tower enabled.
        HintNonpepMetadata(
            hint=complex,
            conf=BeartypeConf(is_pep484_tower=True),
            piths_meta=(
                # Complex number constant.
                HintPithSatisfiedMetadata(2.622 + 575.5j),
                # Floating-point number constant.
                HintPithSatisfiedMetadata(0.8346268),
                # Integer constant.
                HintPithSatisfiedMetadata(1311),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    pith='Park-ed trails',
                    # Match that the exception message raised for this pith
                    # contains...
                    exception_str_match_regexes=(
                        # The type *NOT* satisfied by this object.
                        r'\bcomplex\b',
                        # A newline.
                        r'\n',
                        # A bullet delimiter.
                        r'\*',
                    ),
                    # Match that the exception message raised for this pith
                    # does *NOT* contain...
                    exception_str_not_match_regexes=(
                        # The double-quoted name of this builtin type.
                        r'"complex"',
                    ),
                ),
            ),
        ),

        # ................{ TYPE ~ builtin : fake              }................
        # Fake builtin types (i.e., types that are *NOT* builtin but which
        # nonetheless erroneously masquerade as being builtin), exercising edge
        # cases in @beartype code generation.
        #
        # See also:
        # * The "beartype._data.cls.datacls.TYPES_BUILTIN_FAKE" set.

        # Fake builtin ellipsis type.
        HintNonpepMetadata(
            hint=EllipsisType,
            piths_meta=(
                # Ellipsis singleton.
                HintPithSatisfiedMetadata(...),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'Masterless decree, venomless, which'),
            ),
        ),

        # Fake builtin pure-Python function type.
        HintNonpepMetadata(
            hint=FunctionType,
            piths_meta=(
                # Pure-Python function.
                HintPithSatisfiedMetadata(hints_nonpep484_meta),
                # String constant.
                HintPithUnsatisfiedMetadata('Nomenclature weather‐vanes of'),
            ),
        ),

        # Fake builtin C-based function type.
        HintNonpepMetadata(
            hint=FunctionOrMethodCType,
            piths_meta=(
                # C-based function.
                HintPithSatisfiedMetadata(len),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'Nominally unswain, autodidactic idiocracy, less a'),
            ),
        ),

        # Fake builtin bound method type.
        HintNonpepMetadata(
            hint=MethodBoundInstanceOrClassType,
            piths_meta=(
                # Bound method.
                HintPithSatisfiedMetadata(Class().instance_method),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'ç‐omically gnomical whitebellied burden’s empathy of'),
            ),
        ),

        # Fake builtin module type.
        HintNonpepMetadata(
            hint=ModuleType,
            piths_meta=(
                # Imported module.
                HintPithSatisfiedMetadata(sys),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'Earpiece‐piecemealed, mealy straw headpiece-'),
            ),
        ),

        # Fake builtin "None" singleton type.
        HintNonpepMetadata(
            hint=NoneType,
            piths_meta=(
                # "None" singleton.
                HintPithSatisfiedMetadata(None),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'Earned peace appeasement easements'),
            ),
        ),

        # Fake builtin "NotImplemented" type.
        HintNonpepMetadata(
            hint=NotImplementedType,
            piths_meta=(
                # "NotImplemented" singleton.
                HintPithSatisfiedMetadata(NotImplemented),
                # String constant.
                HintPithUnsatisfiedMetadata('Than'),
            ),
        ),
    ))

    # ..................{ RETURN                             }..................
    # Return this list of all PEP-noncompliant type hint metadata.
    return hints_nonpep_meta
